from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union, cast

from pydantic import AnyHttpUrl, BaseModel, ConfigDict, Field, field_validator

from weaviate.util import _capitalize_first_letter
from weaviate.warnings import _Warnings


class ConsistencyLevel(str, Enum):
    """The consistency levels when writing to Weaviate with replication enabled.

    Attributes:
        ALL: Wait for confirmation of write success from all, `N`, replicas.
        ONE: Wait for confirmation of write success from only one replica.
        QUORUM: Wait for confirmation of write success from a quorum: `N/2+1`, of replicas.
    """

    ALL = "ALL"
    ONE = "ONE"
    QUORUM = "QUORUM"


class DataType(str, Enum):
    """The available primitive data types in Weaviate.

    Attributes:
        TEXT: Text data type.
        TEXT_ARRAY: Text array data type.
        INT: Integer data type.
        INT_ARRAY: Integer array data type.
        BOOL: Boolean data type.
        BOOL_ARRAY: Boolean array data type.
        NUMBER: Number data type.
        NUMBER_ARRAY: Number array data type.
        DATE: Date data type.
        DATE_ARRAY: Date array data type.
        UUID: UUID data type.
        UUID_ARRAY: UUID array data type.
        GEO_COORDINATES: Geo coordinates data type.
        BLOB: Blob data type.
        PHONE_NUMBER: Phone number data type.
        OBJECT: Object data type.
        OBJECT_ARRAY: Object array data type.
    """

    TEXT = "text"
    TEXT_ARRAY = "text[]"
    INT = "int"
    INT_ARRAY = "int[]"
    BOOL = "boolean"
    BOOL_ARRAY = "boolean[]"
    NUMBER = "number"
    NUMBER_ARRAY = "number[]"
    DATE = "date"
    DATE_ARRAY = "date[]"
    UUID = "uuid"
    UUID_ARRAY = "uuid[]"
    GEO_COORDINATES = "geoCoordinates"
    BLOB = "blob"
    PHONE_NUMBER = "phoneNumber"
    OBJECT = "object"
    OBJECT_ARRAY = "object[]"


class _VectorIndexType(str, Enum):
    HNSW = "hnsw"


class Tokenization(str, Enum):
    """The available inverted index tokenization methods for text properties in Weaviate.

    Attributes:
        `WORD`
            Tokenize by word.
        `WHITESPACE`
            Tokenize by whitespace.
        `LOWERCASE`
            Tokenize by lowercase.
        `FIELD`
            Tokenize by field.
    """

    WORD = "word"
    WHITESPACE = "whitespace"
    LOWERCASE = "lowercase"
    FIELD = "field"


class Vectorizer(str, Enum):
    """The available vectorization modules in Weaviate.

    These modules encode binary data into lists of floats called vectors.
    See the [docs](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules) for more details.

    Attributes:
        `NONE`
            No vectorizer.
        `TEXT2VEC_COHERE`
            Weaviate module backed by Cohere text-based embedding models.
        `TEXT2VEC_CONTEXTIONARY`
            Weaviate module backed by Contextionary text-based embedding models.
        `TEXT2VEC_GPT4ALL`
            Weaviate module backed by GPT-4-All text-based embedding models.
        `TEXT2VEC_HUGGINGFACE`
            Weaviate module backed by HuggingFace text-based embedding models.
        `TEXT2VEC_OPENAI`
            Weaviate module backed by OpenAI and Azure-OpenAI text-based embedding models.
        `TEXT2VEC_PALM`
            Weaviate module backed by PaLM text-based embedding models.
        `TEXT2VEC_TRANSFORMERS`
            Weaviate module backed by Transformers text-based embedding models.
        `IMG2VEC_NEURAL`
            Weaviate module backed by a ResNet-50 neural network for images.
        `MULTI2VEC_CLIP`
            Weaviate module backed by a Sentence-BERT CLIP model for images and text.
        `MULTI2VEC_BIND`
            Weaviate module backed by the ImageBind model for images, text, audio, depth, IMU, thermal, and video.
        `REF2VEC_CENTROID`
            Weaviate module backed by a centroid-based model that calculates an object's vectors from its referenced vectors.
    """

    NONE = "none"
    TEXT2VEC_COHERE = "text2vec-cohere"
    TEXT2VEC_CONTEXTIONARY = "text2vec-contextionary"
    TEXT2VEC_GPT4ALL = "text2vec-gpt4all"
    TEXT2VEC_HUGGINGFACE = "text2vec-huggingface"
    TEXT2VEC_OPENAI = "text2vec-openai"
    TEXT2VEC_PALM = "text2vec-palm"
    TEXT2VEC_TRANSFORMERS = "text2vec-transformers"
    IMG2VEC_NEURAL = "img2vec-neural"
    MULTI2VEC_CLIP = "multi2vec-clip"
    MULTI2VEC_BIND = "multi2vec-bind"
    REF2VEC_CENTROID = "ref2vec-centroid"


class GenerativeSearches(str, Enum):
    """The available generative search modules in Weaviate.

    These modules generate text from text-based inputs.
    See the [docs](https://weaviate.io/developers/weaviate/modules/reader-generator-modules) for more details.

    Attributes:
        `OPENAI`
            Weaviate module backed by OpenAI and Azure-OpenAI generative models.
        `COHERE`
            Weaviate module backed by Cohere generative models.
        `PALM`
            Weaviate module backed by PaLM generative models.
    """

    OPENAI = "generative-openai"
    COHERE = "generative-cohere"
    PALM = "generative-palm"


class VectorDistance(str, Enum):
    """Vector similarity distance metric to be used in the `VectorIndexConfig` class.

    To ensure optimal search results, we recommend reviewing whether your model provider advises a
    specific distance metric and following their advice.

    Attributes:
        `COSINE`
            Cosine distance: [reference](https://en.wikipedia.org/wiki/Cosine_similarity)
        `DOT`
            Dot distance: [reference](https://en.wikipedia.org/wiki/Dot_product)
        `L2_SQUARED`
            L2 squared distance: [reference](https://en.wikipedia.org/wiki/Euclidean_distance)
        `HAMMING`
            Hamming distance: [reference](https://en.wikipedia.org/wiki/Hamming_distance)
        `MANHATTAN`
            Manhattan distance: [reference](https://en.wikipedia.org/wiki/Taxicab_geometry)
    """

    COSINE = "cosine"
    DOT = "dot"
    L2_SQUARED = "l2-squared"
    HAMMING = "hamming"
    MANHATTAN = "manhattan"


class StopwordsPreset(str, Enum):
    """Preset stopwords to use in the `Stopwords` class.

    Attributes:
        `EN`
            English stopwords.
        `NONE`
            No stopwords.
    """

    NONE = "none"
    EN = "en"


class PQEncoderType(str, Enum):
    """Type of the PQ encoder.

    Attributes:
        `KMEANS`
            K-means encoder.
        `TILE`
            Tile encoder.
    """

    KMEANS = "kmeans"
    TILE = "tile"


class PQEncoderDistribution(str, Enum):
    """Distribution of the PQ encoder.

    Attributes:
        `LOG_NORMAL`
            Log-normal distribution.
        `NORMAL`
            Normal distribution.
    """

    LOG_NORMAL = "log-normal"
    NORMAL = "normal"


class _ConfigCreateModel(BaseModel):
    model_config = ConfigDict(strict=True)

    def _to_dict(self) -> Dict[str, Any]:
        return cast(dict, self.model_dump(exclude_none=True))


class _ConfigUpdateModel(BaseModel):
    model_config = ConfigDict(strict=True)

    def merge_with_existing(self, schema: Dict[str, Any]) -> Dict[str, Any]:
        for cls_field in self.model_fields:
            val = getattr(self, cls_field)
            if val is None:
                continue
            if isinstance(val, Enum):
                schema[cls_field] = str(val.value)
            elif isinstance(val, (int, float, bool, str, list)):
                schema[cls_field] = val
            else:
                assert isinstance(val, _ConfigUpdateModel)
                schema[cls_field] = val.merge_with_existing(schema[cls_field])
        return schema


class _PQEncoderConfigCreate(_ConfigCreateModel):
    type_: PQEncoderType
    distribution: PQEncoderDistribution

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        ret_dict["type"] = ret_dict.pop("type_")
        return ret_dict


class _PQEncoderConfigUpdate(_ConfigUpdateModel):
    type_: Optional[PQEncoderType]
    distribution: Optional[PQEncoderDistribution]

    def merge_with_existing(self, schema: Dict[str, Any]) -> Dict[str, Any]:
        """Must be done manually since Pydantic does not work well with type and type_.

        Errors shadowing type occur if we want to use type as a field name.
        """
        if self.type_ is not None:
            schema["type"] = str(self.type_.value)
        if self.distribution is not None:
            schema["distribution"] = str(self.distribution.value)
        return schema


class _PQConfigCreate(_ConfigCreateModel):
    bitCompression: bool
    centroids: int
    enabled: bool
    encoder: _PQEncoderConfigCreate
    segments: int
    trainingLimit: int = Field(..., ge=100000)

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        ret_dict["encoder"] = {
            "type": ret_dict.pop("encoder_type"),
            "distribution": ret_dict.pop("encoder_distribution"),
        }
        return ret_dict


class _PQConfigUpdate(_ConfigUpdateModel):
    bitCompression: Optional[bool]
    centroids: Optional[int]
    enabled: Optional[bool]
    segments: Optional[int]
    trainingLimit: Optional[int]
    encoder: Optional[_PQEncoderConfigUpdate]


class _VectorIndexConfigCreate(_ConfigCreateModel):
    cleanupIntervalSeconds: int
    distance: VectorDistance
    dynamicEfMin: int
    dynamicEfMax: int
    dynamicEfFactor: int
    efConstruction: int
    ef: int
    flatSearchCutoff: int
    maxConnections: int
    pq: _PQConfigCreate
    skip: bool
    vectorCacheMaxObjects: int


class _VectorIndexConfigUpdate(_ConfigUpdateModel):
    dynamicEfFactor: Optional[int]
    dynamicEfMin: Optional[int]
    dynamicEfMax: Optional[int]
    ef: Optional[int]
    flatSearchCutoff: Optional[int]
    skip: Optional[bool]
    vectorCacheMaxObjects: Optional[int]
    pq: Optional[_PQConfigUpdate]


class _ShardingConfigCreate(_ConfigCreateModel):
    virtualPerPhysical: int
    desiredCount: int
    actualCount: int
    desiredVirtualCount: int
    actualVirtualCount: int
    key: str = "_id"
    strategy: str = "hash"
    function: str = "murmur3"


class _ReplicationConfigCreate(_ConfigCreateModel):
    factor: int


class _ReplicationConfigUpdate(_ConfigUpdateModel):
    factor: Optional[int]


class _BM25ConfigCreate(_ConfigCreateModel):
    b: float
    k1: float


class _BM25ConfigUpdate(_ConfigUpdateModel):
    b: Optional[float]
    k1: Optional[float]


class _StopwordsCreate(_ConfigCreateModel):
    preset: Optional[StopwordsPreset]
    additions: Optional[List[str]]
    removals: Optional[List[str]]


class _StopwordsUpdate(_ConfigUpdateModel):
    preset: Optional[StopwordsPreset]
    additions: Optional[List[str]]
    removals: Optional[List[str]]


class _InvertedIndexConfigCreate(_ConfigCreateModel):
    bm25: _BM25ConfigCreate
    cleanupIntervalSeconds: int
    indexTimestamps: bool
    indexPropertyLength: bool
    indexNullState: bool
    stopwords: _StopwordsCreate


class _InvertedIndexConfigUpdate(_ConfigUpdateModel):
    bm25: Optional[_BM25ConfigUpdate]
    cleanupIntervalSeconds: Optional[int]
    stopwords: Optional[_StopwordsUpdate]


class _MultiTenancyConfigCreate(_ConfigCreateModel):
    enabled: bool = False


class _MultiTenancyConfigUpdate(_ConfigUpdateModel):
    enabled: Optional[bool] = None


class _GenerativeConfig(_ConfigCreateModel):
    generative: GenerativeSearches


class _GenerativeOpenAIConfigBase(_GenerativeConfig):
    generative: GenerativeSearches = Field(
        default=GenerativeSearches.OPENAI, frozen=True, exclude=True
    )
    frequencyPenaltyProperty: Optional[float]
    presencePenaltyProperty: Optional[float]
    maxTokensProperty: Optional[int]
    temperatureProperty: Optional[float]
    topPProperty: Optional[float]


class _GenerativeOpenAIConfig(_GenerativeOpenAIConfigBase):
    model: Optional[str]


class _GenerativeAzureOpenAIConfig(_GenerativeOpenAIConfigBase):
    resourceName: str
    deploymentId: str


class _GenerativeCohereConfig(_GenerativeConfig):
    generative: GenerativeSearches = Field(
        default=GenerativeSearches.COHERE, frozen=True, exclude=True
    )
    kProperty: Optional[int]
    model: Optional[str]
    maxTokensProperty: Optional[int]
    returnLikelihoodsProperty: Optional[str]
    stopSequencesProperty: Optional[List[str]]
    temperatureProperty: Optional[float]


class _GenerativePaLMConfig(_GenerativeConfig):
    generative: GenerativeSearches = Field(
        default=GenerativeSearches.PALM, frozen=True, exclude=True
    )
    apiEndpoint: Optional[str]
    maxOutputTokens: Optional[int]
    modelId: Optional[str]
    projectId: str
    temperature: Optional[float]
    topK: Optional[int]
    topP: Optional[float]


class _VectorizerConfig(_ConfigCreateModel):
    vectorizer: Vectorizer


class _Generative:
    """Use this factory class to create the correct object for the `generative_config` argument in the `collection.create()` method.

    Each staticmethod provides options specific to the named generative search module in the function's name. Under-the-hood data validation steps
    will ensure that any mis-specifications will be caught before the request is sent to Weaviate.
    """

    @staticmethod
    def openai(
        model: Optional[str] = None,
        frequency_penalty: Optional[float] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
    ) -> _GenerativeConfig:
        return _GenerativeOpenAIConfig(
            frequencyPenaltyProperty=frequency_penalty,
            maxTokensProperty=max_tokens,
            model=model,
            presencePenaltyProperty=presence_penalty,
            temperatureProperty=temperature,
            topPProperty=top_p,
        )

    @staticmethod
    def azure_openai(
        resource_name: str,
        deployment_id: str,
        frequency_penalty: Optional[float] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
    ) -> _GenerativeConfig:
        return _GenerativeAzureOpenAIConfig(
            deploymentId=deployment_id,
            frequencyPenaltyProperty=frequency_penalty,
            maxTokensProperty=max_tokens,
            presencePenaltyProperty=presence_penalty,
            resourceName=resource_name,
            temperatureProperty=temperature,
            topPProperty=top_p,
        )

    @staticmethod
    def cohere(
        model: Optional[str] = None,
        k: Optional[int] = None,
        max_tokens: Optional[int] = None,
        return_likelihoods: Optional[str] = None,
        stop_sequences: Optional[List[str]] = None,
        temperature: Optional[float] = None,
    ) -> _GenerativeConfig:
        return _GenerativeCohereConfig(
            kProperty=k,
            maxTokensProperty=max_tokens,
            model=model,
            returnLikelihoodsProperty=return_likelihoods,
            stopSequencesProperty=stop_sequences,
            temperatureProperty=temperature,
        )

    @staticmethod
    def palm(
        project_id: str,
        api_endpoint: Optional[str] = None,
        max_output_tokens: Optional[int] = None,
        model_id: Optional[str] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
    ) -> _GenerativeConfig:
        return _GenerativePaLMConfig(
            apiEndpoint=api_endpoint,
            maxOutputTokens=max_output_tokens,
            modelId=model_id,
            projectId=project_id,
            temperature=temperature,
            topK=top_k,
            topP=top_p,
        )


class _Text2VecContextionaryConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(
        default=Vectorizer.TEXT2VEC_CONTEXTIONARY, frozen=True, exclude=True
    )
    vectorizeClassName: bool


CohereModel = Literal[
    "embed-multilingual-v2.0",
    "small",
    "medium",
    "large",
    "multilingual-22-12",
    "embed-english-v2.0",
    "embed-english-light-v2.0",
]
CohereTruncation = Literal["RIGHT", "NONE"]


class _Text2VecCohereConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(default=Vectorizer.TEXT2VEC_COHERE, frozen=True, exclude=True)
    model: Optional[CohereModel]
    truncate: Optional[CohereTruncation]


class _Text2VecHuggingFaceConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(
        default=Vectorizer.TEXT2VEC_HUGGINGFACE, frozen=True, exclude=True
    )
    model: Optional[str]
    passageModel: Optional[str]
    queryModel: Optional[str]
    endpointURL: Optional[str]
    waitForModel: Optional[bool]
    useGPU: Optional[bool]
    useCache: Optional[bool]

    def validate_mutually_exclusive_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if "passageModel" in values and "queryModel" not in values:
            raise ValueError("Must specify query_model when specifying passage_model")
        if "queryModel" in values and "passageModel" not in values:
            raise ValueError("Must specify passage_model when specifying query_model")
        if "model" in values and any(["passageModel" in values, "queryModel" in values]):
            raise ValueError(
                "Can only specify model alone or passage_model and query_model together"
            )
        if (
            any(["passageModel" in values, "queryModel" in values, "model" in values])
            and "endpointURL" in values
        ):
            _Warnings.text2vec_huggingface_endpoint_url_and_model_set_together()
        if all(
            [
                "passageModel" not in values,
                "queryModel" not in values,
                "model" not in values,
                "endpointURL" not in values,
            ]
        ):
            raise ValueError(
                "Must specify at least one of model, passage_model & query_model, or endpoint_url"
            )
        return values

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        options = {}
        if self.waitForModel is not None:
            options["waitForModel"] = ret_dict.pop("waitForModel")
        if self.useGPU is not None:
            options["useGPU"] = ret_dict.pop("useGPU")
        if self.useCache is not None:
            options["useCache"] = ret_dict.pop("useCache")
        if len(options) > 0:
            ret_dict["options"] = options
        return ret_dict


OpenAIModel = Literal["ada", "babbage", "curie", "davinci"]
OpenAIType = Literal["text", "code"]


class _Text2VecOpenAIConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(default=Vectorizer.TEXT2VEC_OPENAI, frozen=True, exclude=True)
    model: Optional[OpenAIModel]
    modelVersion: Optional[str]
    type_: Optional[OpenAIType]
    vectorizeClassName: bool

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        if self.type_ is not None:
            ret_dict["type"] = ret_dict.pop("type_")
        return ret_dict


class _Text2VecAzureOpenAIConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(default=Vectorizer.TEXT2VEC_OPENAI, frozen=True, exclude=True)
    resourceName: str
    deploymentId: str


class _Text2VecPalmConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(default=Vectorizer.TEXT2VEC_PALM, frozen=True, exclude=True)
    projectId: str
    apiEndpoint: Optional[AnyHttpUrl]
    modelId: Optional[str]
    vectorizeClassName: bool

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        if self.apiEndpoint is not None:
            ret_dict["apiEndpoint"] = str(self.apiEndpoint)
        return ret_dict


class _Text2VecTransformersConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(
        default=Vectorizer.TEXT2VEC_TRANSFORMERS, frozen=True, exclude=True
    )
    poolingStrategy: Literal["masked_mean", "cls"]
    vectorizeClassName: bool


class _Text2VecGPT4AllConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(default=Vectorizer.TEXT2VEC_GPT4ALL, frozen=True, exclude=True)
    vectorizeClassName: bool


class _Img2VecNeuralConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(default=Vectorizer.IMG2VEC_NEURAL, frozen=True, exclude=True)
    imageFields: List[str]


class Multi2VecField(BaseModel):
    """Use this class when defining the fields to use in the `Multi2VecClip` and `Multi2VecBind` vectorizers."""

    name: str
    weight: Optional[float] = Field(default=None, exclude=True)


class _Multi2VecBase(_VectorizerConfig):
    imageFields: Optional[List[Multi2VecField]]
    textFields: Optional[List[Multi2VecField]]
    vectorizeClassName: bool

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        ret_dict["weights"] = {}
        for cls_field in self.model_fields:
            val = getattr(self, cls_field)
            if "Fields" in cls_field and val is not None:
                val = cast(List[Multi2VecField], val)
                ret_dict[cls_field] = [field.name for field in val]
                weights = [field.weight for field in val if field.weight is not None]
                if len(weights) > 0:
                    ret_dict["weights"][cls_field] = weights
        if len(ret_dict["weights"]) == 0:
            del ret_dict["weights"]
        return ret_dict


class _Multi2VecClipConfig(_Multi2VecBase):
    vectorizer: Vectorizer = Field(default=Vectorizer.MULTI2VEC_CLIP, frozen=True, exclude=True)


class _Multi2VecBindConfig(_Multi2VecBase):
    vectorizer: Vectorizer = Field(default=Vectorizer.MULTI2VEC_BIND, frozen=True, exclude=True)
    audioFields: Optional[List[Multi2VecField]]
    depthFields: Optional[List[Multi2VecField]]
    IMUFields: Optional[List[Multi2VecField]]
    thermalFields: Optional[List[Multi2VecField]]
    videoFields: Optional[List[Multi2VecField]]


class _Ref2VecCentroidConfig(_VectorizerConfig):
    vectorizer: Vectorizer = Field(default=Vectorizer.REF2VEC_CENTROID, frozen=True, exclude=True)
    referenceProperties: List[str]
    method: Literal["mean"]


class _Vectorizer:
    """Use this factory class to create the correct object for the `vectorizer_config` argument in the `collection.create()` method.

    Each staticmethod provides options specific to the named vectorizer in the function's name. Under-the-hood data validation steps
    will ensure that any mis-specifications will be caught before the request is sent to Weaviate.
    """

    @staticmethod
    def none() -> _VectorizerConfig:
        """Create a `VectorizerConfig` object with the vectorizer set to `Vectorizer.NONE`."""
        return _VectorizerConfig(vectorizer=Vectorizer.NONE)

    @staticmethod
    def img2vec_neural(
        image_fields: List[str],
    ) -> _VectorizerConfig:
        """Create a `Img2VecNeuralConfig` object for use when vectorizing using the `img2vec-neural` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/img2vec-neural)
        for detailed usage.

        Arguments:
            `image_fields`
                The image fields to use. This is a required field and must match the property fields
                of the collection that are defined as `DataType.BLOB`.

        Raises:
            `pydantic.ValidationError` if `image_fields` is not a `list`.
        """
        return _Img2VecNeuralConfig(imageFields=image_fields)

    @staticmethod
    def multi2vec_clip(
        image_fields: Optional[List[Multi2VecField]] = None,
        text_fields: Optional[List[Multi2VecField]] = None,
        vectorize_class_name: bool = True,
    ) -> _VectorizerConfig:
        """Create a `Multi2VecClipConfig` object for use when vectorizing using the `multi2vec-clip` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/multi2vec-clip)
        for detailed usage.

        Arguments:
            `image_fields`
                The image fields to use in vectorization.
            `text_fields`
                The text fields to use in vectorization.
            `vectorize_class_name`
                Whether to vectorize the class name. Defaults to `True`.

        Raises:
            `pydantic.ValidationError` if `image_fields` or `text_fields` are not `None` or a `list`.
        """
        return _Multi2VecClipConfig(
            imageFields=image_fields,
            textFields=text_fields,
            vectorizeClassName=vectorize_class_name,
        )

    @staticmethod
    def multi2vec_bind(
        audio_fields: Optional[List[Multi2VecField]] = None,
        depth_fields: Optional[List[Multi2VecField]] = None,
        image_fields: Optional[List[Multi2VecField]] = None,
        imu_fields: Optional[List[Multi2VecField]] = None,
        text_fields: Optional[List[Multi2VecField]] = None,
        thermal_fields: Optional[List[Multi2VecField]] = None,
        video_fields: Optional[List[Multi2VecField]] = None,
        vectorize_class_name: bool = True,
    ) -> _VectorizerConfig:
        """Create a `Multi2VecClipConfig` object for use when vectorizing using the `multi2vec-clip` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/multi2vec-bind)
        for detailed usage.

        Arguments:
            `audio_fields`
                The audio fields to use in vectorization.
            `depth_fields`
                The depth fields to use in vectorization.
            `image_fields`
                The image fields to use in vectorization.
            `imu_fields`
                The IMU fields to use in vectorization.
            `text_fields`
                The text fields to use in vectorization.
            `thermal_fields`
                The thermal fields to use in vectorization.
            `video_fields`
                The video fields to use in vectorization.
            `vectorize_class_name`
                Whether to vectorize the class name. Defaults to `True`.

        Raises:
            `pydantic.ValidationError` if any of the `*_fields` are not `None` or a `list`.
        """
        return _Multi2VecBindConfig(
            audioFields=audio_fields,
            depthFields=depth_fields,
            imageFields=image_fields,
            IMUFields=imu_fields,
            textFields=text_fields,
            thermalFields=thermal_fields,
            videoFields=video_fields,
            vectorizeClassName=vectorize_class_name,
        )

    @staticmethod
    def ref2vec_centroid(
        reference_properties: List[str],
        method: Literal["mean"] = "mean",
    ) -> _VectorizerConfig:
        """Create a `Ref2VecCentroidConfig` object for use when vectorizing using the `ref2vec-centroid` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/ref2vec-centroid)
        for detailed usage.

        Arguments:
            `reference_properties`
                The reference properties to use in vectorization, REQUIRED.
            `method`
                The method to use in vectorization. Defaults to `mean`.

        Raises:
            `pydantic.ValidationError` if `reference_properties` is not a `list`.
        """
        return _Ref2VecCentroidConfig(
            referenceProperties=reference_properties,
            method=method,
        )

    @staticmethod
    def text2vec_azure_openai(resource_name: str, deployment_id: str) -> _VectorizerConfig:
        """Create a `Text2VecAzureOpenAIConfig` object for use when vectorizing using the `text2vec-azure-openai` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-azure-openai)
        for detailed usage.

        Arguments:
            `resource_name`
                The resource name to use, REQUIRED.
            `deployment_id`
                The deployment ID to use, REQUIRED.
            `vectorize_class_name`
                Whether to vectorize the class name. Defaults to `True`.

        Raises:
            `pydantic.ValidationError` if `resource_name` or `deployment_id` are not `str`.
        """
        return _Text2VecAzureOpenAIConfig(resourceName=resource_name, deploymentId=deployment_id)

    @staticmethod
    def text2vec_contextionary(vectorize_class_name: bool = True) -> _VectorizerConfig:
        """Create a `Text2VecContextionaryConfig` object for use when vectorizing using the `text2vec-contextionary` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-contextionary)
        for detailed usage.

        Arguments:
            vectorize_class_name: Whether to vectorize the class name. Defaults to `True`.

        Raises:
            pydantic.ValidationError` if `vectorize_class_name` is not a `bool`.
        """
        return _Text2VecContextionaryConfig(vectorizeClassName=vectorize_class_name)

    @staticmethod
    def text2vec_cohere(
        model: Optional[CohereModel] = None,
        truncate: Optional[CohereTruncation] = None,
    ) -> _VectorizerConfig:
        """Create a `Text2VecCohereConfig` object for use when vectorizing using the `text2vec-cohere` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-cohere)
        for detailed usage.

        Arguments:
            `model`
                The model to use. Defaults to `None`. If `None`, the default model is used.
            `truncate`
                The truncation strategy to use. Defaults to `None`. If `None`, the default truncation strategy is used.

        Raises:
            `pydantic.ValidationError` if `model` or `truncate` are not valid values from the `CohereModel` and `CohereTruncate` types.
        """
        return _Text2VecCohereConfig(model=model, truncate=truncate)

    @staticmethod
    def text2vec_gpt4all(
        vectorize_class_name: bool = True,
    ) -> _VectorizerConfig:
        """Create a `Text2VecGPT4AllConfig` object for use when vectorizing using the `text2vec-gpt4all` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-gpt4all)
        for detailed usage.

        Arguments:
            `vectorize_class_name`
                Whether to vectorize the class name. Defaults to `True`.

        Raises:
            `pydantic.ValidationError` if `vectorize_class_name` is not a `bool`.
        """
        return _Text2VecGPT4AllConfig(vectorizeClassName=vectorize_class_name)

    @staticmethod
    def text2vec_huggingface(
        model: Optional[str] = None,
        passage_model: Optional[str] = None,
        query_model: Optional[str] = None,
        endpoint_url: Optional[str] = None,
        wait_for_model: Optional[bool] = None,
        use_gpu: Optional[bool] = None,
        use_cache: Optional[bool] = None,
    ) -> _VectorizerConfig:
        """Create a `Text2VecHuggingFaceConfig` object for use when vectorizing using the `text2vec-huggingface` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-huggingface)
        for detailed usage.

        Arguments:
            `model`
                The model to use. Defaults to `None`.
            `passage_model`
                The passage model to use. Defaults to `None`.
            `query_model`
                The query model to use. Defaults to `None`.
            `endpoint_url`
                The endpoint URL to use. Defaults to `None`.
            `wait_for_model`
                Whether to wait for the model to be loaded. Defaults to `None`.
            `use_gpu`
                Whether to use the GPU. Defaults to `None`.
            `use_cache`
                Whether to use the cache. Defaults to `None`.

        Raises:
            `pydantic.ValidationError` if the arguments passed to the function are invalid.
                It is important to note that some of these variables are mutually exclusive.
                    See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-huggingface) for more details.
        """
        return _Text2VecHuggingFaceConfig(
            model=model,
            passageModel=passage_model,
            queryModel=query_model,
            endpointURL=endpoint_url,
            waitForModel=wait_for_model,
            useGPU=use_gpu,
            useCache=use_cache,
        )

    @staticmethod
    def text2vec_openai(
        model: Optional[OpenAIModel] = None,
        model_version: Optional[str] = None,
        type_: Optional[OpenAIType] = None,
        vectorize_class_name: bool = True,
    ) -> _VectorizerConfig:
        """Create a `Text2VecOpenAIConfig` object for use when vectorizing using the `text2vec-openai` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-openai)
        for detailed usage.

        Arguments:
            `model`
                The model to use. Defaults to `None`. If `None`, the default model is used.
            `model_version`
                The model version to use. Defaults to `None`.
            `type_`
                The type of model to use. Defaults to `None`. If `None`, the default type is used.
            `vectorize_class_name`
                Whether to vectorize the class name. Defaults to `True`.

        Raises:
            `pydantic.ValidationError` if `model` or `type_` are not valid values from the `OpenAIModel` and `OpenAIType` types.
        """
        return _Text2VecOpenAIConfig(
            model=model,
            modelVersion=model_version,
            type_=type_,
            vectorizeClassName=vectorize_class_name,
        )

    @staticmethod
    def text2vec_palm(
        project_id: str,
        api_endpoint: Optional[AnyHttpUrl] = None,
        model_id: Optional[str] = None,
        vectorize_class_name: bool = True,
    ) -> _VectorizerConfig:
        """Create a `Text2VecPalmConfig` object for use when vectorizing using the `text2vec-palm` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-palm)
        for detailed usage.

        Arguments:
            `project_id`
                The project ID to use, REQUIRED.
            `api_endpoint`
                The API endpoint to use. Defaults to `None`.
            `model_id`
                The model ID to use. Defaults to `None`.
            `vectorize_class_name`
                Whether to vectorize the class name. Defaults to `True`.

        Raises:
            `pydantic.ValidationError` if `api_endpoint` is not a valid URL.
        """
        return _Text2VecPalmConfig(
            projectId=project_id,
            apiEndpoint=api_endpoint,
            modelId=model_id,
            vectorizeClassName=vectorize_class_name,
        )

    @staticmethod
    def text2vec_transformers(
        pooling_strategy: Literal["masked_mean", "cls"] = "masked_mean",
        vectorize_class_name: bool = True,
    ) -> _VectorizerConfig:
        """Create a `Text2VecTransformersConfig` object for use when vectorizing using the `text2vec-transformers` model.

        See the [documentation](https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules/text2vec-transformers)
        for detailed usage.

        Arguments:
            `pooling_strategy`
                The pooling strategy to use. Defaults to `masked_mean`.
            `vectorize_class_name`
                Whether to vectorize the class name. Defaults to `True`.

        Raises:
            `pydantic.ValidationError` if `pooling_strategy` is not a valid value from the `PoolingStrategy` type.
        """
        return _Text2VecTransformersConfig(
            poolingStrategy=pooling_strategy,
            vectorizeClassName=vectorize_class_name,
        )


class _CollectionConfigCreateBase(_ConfigCreateModel):
    description: Optional[str] = Field(default=None)
    invertedIndexConfig: Optional[_InvertedIndexConfigCreate] = Field(
        default=None, alias="inverted_index_config"
    )
    multiTenancyConfig: Optional[_MultiTenancyConfigCreate] = Field(
        default=None, alias="multi_tenancy_config"
    )
    replicationConfig: Optional[_ReplicationConfigCreate] = Field(
        default=None, alias="replication_config"
    )
    shardingConfig: Optional[_ShardingConfigCreate] = Field(default=None, alias="sharding_config")
    vectorIndexConfig: Optional[_VectorIndexConfigCreate] = Field(
        default=None, alias="vector_index_config"
    )
    vectorIndexType: _VectorIndexType = Field(
        default=_VectorIndexType.HNSW, alias="vector_index_type"
    )
    moduleConfig: _VectorizerConfig = Field(default=_Vectorizer.none(), alias="vectorizer_config")
    generativeSearch: Optional[_GenerativeConfig] = Field(default=None, alias="generative_config")

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict: Dict[str, Any] = {}

        for cls_field in self.model_fields:
            val = getattr(self, cls_field)
            if cls_field in ["name", "model", "properties"] or val is None:
                continue
            if isinstance(val, Enum):
                ret_dict[cls_field] = str(val.value)
            elif isinstance(val, (bool, float, str, int)):
                ret_dict[cls_field] = str(val)
            elif isinstance(val, _GenerativeConfig):
                self.__add_to_module_config(ret_dict, val.generative.value, val._to_dict())
            elif isinstance(val, _VectorizerConfig):
                ret_dict["vectorizer"] = val.vectorizer.value
                if val.vectorizer != Vectorizer.NONE:
                    self.__add_to_module_config(ret_dict, val.vectorizer.value, val._to_dict())
            else:
                assert isinstance(val, _ConfigCreateModel)
                ret_dict[cls_field] = val._to_dict()
        if self.moduleConfig is None:
            ret_dict["vectorizer"] = Vectorizer.NONE.value
        return ret_dict

    @staticmethod
    def __add_to_module_config(
        return_dict: Dict[str, Any], addition_key: str, addition_val: Dict[str, Any]
    ) -> None:
        if "moduleConfig" not in return_dict:
            return_dict["moduleConfig"] = {addition_key: addition_val}
        else:
            return_dict["moduleConfig"][addition_key] = addition_val


class _CollectionConfigUpdate(_ConfigUpdateModel):
    description: Optional[str] = Field(default=None)
    invertedIndexConfig: Optional[_InvertedIndexConfigUpdate] = Field(
        default=None, alias="inverted_index_config"
    )
    replicationConfig: Optional[_ReplicationConfigUpdate] = Field(
        default=None, alias="replication_config"
    )
    vectorIndexConfig: Optional[_VectorIndexConfigUpdate] = Field(
        default=None, alias="vector_index_config"
    )


@dataclass
class _BM25Config:
    b: float
    k1: float


@dataclass
class _StopwordsConfig:
    preset: StopwordsPreset
    additions: Optional[List[str]]
    removals: Optional[List[str]]


@dataclass
class _InvertedIndexConfig:
    bm25: _BM25Config
    cleanup_interval_seconds: int
    index_null_state: bool
    index_property_length: bool
    index_timestamps: bool
    stopwords: _StopwordsConfig


@dataclass
class _MultiTenancyConfig:
    enabled: bool


@dataclass
class _ReferenceDataType:
    target_collection: str


@dataclass
class _ReferenceDataTypeMultiTarget:
    target_collections: List[str]


@dataclass
class _Property:
    data_type: Union[DataType, _ReferenceDataType, _ReferenceDataTypeMultiTarget]
    description: Optional[str]
    index_filterable: bool
    index_searchable: bool
    name: str
    tokenization: Optional[Tokenization]

    def to_weaviate_dict(self) -> Dict[str, Any]:
        if isinstance(self.data_type, DataType):
            data_type = [self.data_type.value]
        elif isinstance(self.data_type, _ReferenceDataType):
            data_type = [self.data_type.target_collection]
        else:
            data_type = self.data_type.target_collections
        return {
            "dataType": data_type,
            "description": self.description,
            "indexFilterable": self.index_filterable,
            "indexVector": self.index_searchable,
            "name": self.name,
            "tokenizer": self.tokenization.value if self.tokenization else None,
        }


@dataclass
class _ReplicationConfig:
    factor: int


@dataclass
class _ShardingConfig:
    virtual_per_physical: int
    desired_count: int
    actual_count: int
    desired_virtual_count: int
    actual_virtual_count: int
    key: str
    strategy: str
    function: str


@dataclass
class _PQEncoderConfig:
    type_: PQEncoderType
    distribution: PQEncoderDistribution


@dataclass
class _PQConfig:
    enabled: bool
    bit_compression: bool
    segments: int
    centroids: int
    training_limit: int
    encoder: _PQEncoderConfig


@dataclass
class _VectorIndexConfig:
    cleanup_interval_seconds: int
    distance_metric: VectorDistance
    dynamic_ef_min: int
    dynamic_ef_max: int
    dynamic_ef_factor: int
    ef: int
    ef_construction: int
    flat_search_cutoff: int
    max_connections: int
    pq: _PQConfig
    skip: bool
    vector_cache_max_objects: int


@dataclass
class _CollectionConfig:
    name: str
    description: Optional[str]
    inverted_index_config: _InvertedIndexConfig
    multi_tenancy_config: _MultiTenancyConfig
    properties: List[_Property]
    replication_config: _ReplicationConfig
    sharding_config: _ShardingConfig
    vector_index_config: _VectorIndexConfig
    vector_index_type: _VectorIndexType
    vectorizer: Vectorizer


@dataclass
class _CollectionConfigSimple:
    name: str
    description: Optional[str]
    properties: List[_Property]
    vectorizer: Vectorizer


@dataclass
class _ShardStatus:
    name: str
    status: Literal["READONLY", "READY", "INDEXING"]
    vector_queue_size: int


# class PropertyConfig(ConfigCreateModel):
#     indexFilterable: Optional[bool] = Field(None, alias="index_filterable")
#     indexSearchable: Optional[bool] = Field(None, alias="index_searchable")
#     tokenization: Optional[Tokenization] = None
#     description: Optional[str] = None
#     moduleConfig: Optional[ModuleConfig] = Field(None, alias="module_config")


@dataclass
class PropertyConfig:  # noqa
    index_filterable: Optional[bool] = None
    index_searchable: Optional[bool] = None
    tokenization: Optional[Tokenization] = None
    description: Optional[str] = None
    vectorizer_config: Optional[_VectorizerConfig] = None

    # tmp solution. replace with a pydantic BaseModel, see bugreport: https://github.com/pydantic/pydantic/issues/6948
    # bugreport was closed as not planned :( so dataclasses must stay
    def _to_dict(self) -> Dict[str, Any]:
        return {
            "indexFilterable": self.index_filterable,
            "indexSearchable": self.index_searchable,
            "tokenization": self.tokenization,
            "description": self.description,
            "moduleConfig": self.vectorizer_config,
        }


class Property(_ConfigCreateModel):
    """This class defines the structure of a data property that a collection can have within Weaviate.

    Attributes:
        `name`
            The name of the property, REQUIRED.
        `data_type`
            The data type of the property, REQUIRED.
        `index_filterable`
            Whether the property should be filterable in the inverted index.
        `index_searchable`
            Whether the property should be searchable in the inverted index.
        `description`
            A description of the property.
        `skip_vectorization`
            Whether to skip vectorization of the property. Defaults to `False`.
        `tokenization`
            The tokenization method to use for the inverted index. Defaults to `None`.
        `vectorize_property_name`
            Whether to vectorize the property name. Defaults to `True`.
    """

    name: str
    dataType: DataType = Field(default=..., alias="data_type")
    indexFilterable: Optional[bool] = Field(default=None, alias="index_filterable")
    indexSearchable: Optional[bool] = Field(default=None, alias="index_searchable")
    description: Optional[str] = Field(default=None)
    nestedProperties: Optional[Union["Property", List["Property"]]] = Field(
        default=None, alias="nested_properties"
    )
    skip_vectorization: bool = Field(default=False)
    tokenization: Optional[Tokenization] = Field(default=None)
    vectorize_property_name: bool = Field(default=True)

    @field_validator("name")
    def _check_name(cls, v: str) -> str:
        if v in ["id", "vector"]:
            raise ValueError(f"Property name '{v}' is reserved and cannot be used")
        return v

    def _to_dict(self, vectorizer: Optional[Vectorizer] = None) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        ret_dict["dataType"] = [ret_dict["dataType"]]
        if vectorizer is not None and vectorizer != Vectorizer.NONE:
            ret_dict["moduleConfig"] = {
                vectorizer.value: {
                    "skip": self.skip_vectorization,
                    "vectorizePropertyName": self.vectorize_property_name,
                }
            }
        del ret_dict["skip_vectorization"]
        del ret_dict["vectorize_property_name"]
        if self.nestedProperties is not None:
            ret_dict["nestedProperties"] = (
                [prop._to_dict() for prop in self.nestedProperties]
                if isinstance(self.nestedProperties, list)
                else [self.nestedProperties._to_dict()]
            )
        return ret_dict


class _ReferencePropertyBase(_ConfigCreateModel):
    name: str

    @field_validator("name")
    def check_name(cls, v: str) -> str:
        if v in ["id", "vector"]:
            raise ValueError(f"Property name '{v}' is reserved and cannot be used")
        return v


class ReferenceProperty(_ReferencePropertyBase):
    """This class defines properties that are cross references to a single target collection.

    Use this class when you want to create a cross-reference in the collection's config that is capable
    of having only cross-references to a single other collection.

    Attributes:
        `name`
            The name of the property, REQUIRED.
        `target_collection`
            The name of the target collection, REQUIRED.
    """

    target_collection: str

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        ret_dict["dataType"] = [_capitalize_first_letter(self.target_collection)]
        del ret_dict["target_collection"]
        return ret_dict


class ReferencePropertyMultiTarget(_ReferencePropertyBase):
    """This class defines properties that are cross references to multiple target collections.

    Use this class when you want to create a cross-reference in the collection's config that is capable
    of having cross-references to multiple other collections at once.

    Attributes:
        `name`
            The name of the property, REQUIRED.
        `target_collections`
            The names of the target collections, REQUIRED.
    """

    target_collections: List[str]

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        ret_dict["dataType"] = [
            _capitalize_first_letter(target) for target in self.target_collections
        ]
        del ret_dict["target_collections"]
        return ret_dict


PropertyType = Union[Property, ReferenceProperty, ReferencePropertyMultiTarget]


class _CollectionConfigCreate(_CollectionConfigCreateBase):
    name: str
    properties: Optional[List[Union[Property, _ReferencePropertyBase]]] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        self.name = _capitalize_first_letter(self.name)

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()

        ret_dict["class"] = self.name

        if self.properties is not None:
            ret_dict["properties"] = [
                prop._to_dict(self.moduleConfig.vectorizer)
                if isinstance(prop, Property)
                else prop._to_dict()
                for prop in self.properties
            ]

        return ret_dict


class Configure:
    """Use this factory class to generate the correct object for use when using the `collection.create()` method. E.g., `.multi_tenancy()` will return a `MultiTenancyConfigCreate` object to be used in the `multi_tenancy_config` argument.

    Each class method provides options specific to the named configuration type in the function's name. Under-the-hood data validation steps
    will ensure that any mis-specifications are caught before the request is sent to Weaviate.
    """

    Generative = _Generative
    Vectorizer = _Vectorizer

    @staticmethod
    def inverted_index(
        bm25_b: float = 0.75,
        bm25_k1: float = 1.2,
        cleanup_interval_seconds: int = 60,
        index_timestamps: bool = False,
        index_property_length: bool = False,
        index_null_state: bool = False,
        stopwords_preset: Optional[StopwordsPreset] = None,
        stopwords_additions: Optional[List[str]] = None,
        stopwords_removals: Optional[List[str]] = None,
    ) -> _InvertedIndexConfigCreate:
        """Create an `InvertedIndexConfigCreate` object to be used when defining the configuration of the keyword searching algorithm of Weaviate.

        Arguments:
            See [the docs](https://weaviate.io/developers/weaviate/configuration/indexes#configure-the-inverted-index) for details!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _InvertedIndexConfigCreate(
            bm25=_BM25ConfigCreate(b=bm25_b, k1=bm25_k1),
            cleanupIntervalSeconds=cleanup_interval_seconds,
            indexTimestamps=index_timestamps,
            indexPropertyLength=index_property_length,
            indexNullState=index_null_state,
            stopwords=_StopwordsCreate(
                preset=stopwords_preset,
                additions=stopwords_additions,
                removals=stopwords_removals,
            ),
        )

    @staticmethod
    def multi_tenancy(enabled: bool = False) -> _MultiTenancyConfigCreate:
        """Create a `MultiTenancyConfigCreate` object to be used when defining the multi-tenancy configuration of Weaviate.

        Arguments:
            `enabled`
                Whether multi-tenancy is enabled. Defaults to `False`.
        """
        return _MultiTenancyConfigCreate(enabled=enabled)

    @staticmethod
    def replication(factor: int = 1) -> _ReplicationConfigCreate:
        """Create a `ReplicationConfigCreate` object to be used when defining the replication configuration of Weaviate.

        Arguments:
            `factor`
                The replication factor. Defaults to `1`.
        """
        return _ReplicationConfigCreate(factor=factor)

    @staticmethod
    def sharding(
        virtual_per_physical: int = 128,
        desired_count: int = 1,
        actual_count: int = 1,
        desired_virtual_count: int = 128,
        actual_virtual_count: int = 128,
    ) -> _ShardingConfigCreate:
        """Create a `ShardingConfigCreate` object to be used when defining the sharding configuration of Weaviate.

        NOTE: You can only use one of Sharding or Replication, not both.

        See [the docs](https://weaviate.io/developers/weaviate/concepts/replication-architecture#replication-vs-sharding) for more details.

        Arguments:
            `virtual_per_physical`
                The number of virtual shards per physical shard. Defaults to `128`.
            `desired_count`
                The desired number of physical shards. Defaults to `1`.
            `actual_count`
                The actual number of physical shards. Defaults to `1`.
            `desired_virtual_count`
                The desired number of virtual shards. Defaults to `128`.
            `actual_virtual_count`
                The actual number of virtual shards. Defaults to `128`.
        """
        return _ShardingConfigCreate(
            virtualPerPhysical=virtual_per_physical,
            desiredCount=desired_count,
            actualCount=actual_count,
            desiredVirtualCount=desired_virtual_count,
            actualVirtualCount=actual_virtual_count,
        )

    @staticmethod
    def vector_index(
        cleanup_interval_seconds: int = 300,
        distance_metric: VectorDistance = VectorDistance.COSINE,
        dynamic_ef_factor: int = 8,
        dynamic_ef_max: int = 500,
        dynamic_ef_min: int = 100,
        ef: int = -1,
        ef_construction: int = 128,
        flat_search_cutoff: int = 40000,
        max_connections: int = 64,
        pq_bit_compression: bool = False,
        pq_centroids: int = 256,
        pq_enabled: bool = False,
        pq_encoder_distribution: PQEncoderDistribution = PQEncoderDistribution.LOG_NORMAL,
        pq_encoder_type: PQEncoderType = PQEncoderType.KMEANS,
        pq_segments: int = 0,
        pq_training_limit: int = 100000,
        skip: bool = False,
        vector_cache_max_objects: int = 1000000000000,
    ) -> _VectorIndexConfigCreate:
        """Create a `_VectorIndexConfigCreate` object to be used when defining the vector index configuration of Weaviate.

        Use this method when defining the `vector_index_config` argument in `collection.create()`.

        Arguments:
            See [the docs](https://weaviate.io/developers/weaviate/configuration/indexes#how-to-configure-hnsw) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _VectorIndexConfigCreate(
            cleanupIntervalSeconds=cleanup_interval_seconds,
            distance=distance_metric,
            dynamicEfMin=dynamic_ef_min,
            dynamicEfMax=dynamic_ef_max,
            dynamicEfFactor=dynamic_ef_factor,
            efConstruction=ef_construction,
            ef=ef,
            flatSearchCutoff=flat_search_cutoff,
            maxConnections=max_connections,
            pq=_PQConfigCreate(
                bitCompression=pq_bit_compression,
                centroids=pq_centroids,
                enabled=pq_enabled,
                encoder=_PQEncoderConfigCreate(
                    type_=pq_encoder_type,
                    distribution=pq_encoder_distribution,
                ),
                segments=pq_segments,
                trainingLimit=pq_training_limit,
            ),
            skip=skip,
            vectorCacheMaxObjects=vector_cache_max_objects,
        )

    @staticmethod
    def vector_index_type() -> _VectorIndexType:
        """Create a `_VectorIndexType` object to be used when defining the vector index type of Weaviate.

        Use this method when defining the `vector_index_type` argument in `collection.create()`.
        """
        return _VectorIndexType.HNSW


class ConfigureUpdate:
    """Use this factory class to generate the correct `xxxConfig` object for use when using the `collection.update()` method.

    Each staticmethod provides options specific to the named configuration type in the function's name. Under-the-hood data validation steps
    will ensure that any mis-specifications are caught before the request is sent to Weaviate. Only those configurations that are mutable are
    available in this class. If you wish to update the configuration of an immutable aspect of your collection then you will have to delete
    the collection and re-create it with the new configuration.
    """

    @staticmethod
    def inverted_index(
        bm25_b: Optional[float] = None,
        bm25_k1: Optional[float] = None,
        cleanup_interval_seconds: Optional[int] = None,
        stopwords_additions: Optional[List[str]] = None,
        stopwords_preset: Optional[StopwordsPreset] = None,
        stopwords_removals: Optional[List[str]] = None,
    ) -> _InvertedIndexConfigUpdate:
        """Create an `InvertedIndexConfigUpdate` object.

        Use this method when defining the `inverted_index_config` argument in `collection.update()`.

        Arguments:
            See [the docs](https://weaviate.io/developers/weaviate/configuration/indexes#configure-the-inverted-index) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _InvertedIndexConfigUpdate(
            bm25=_BM25ConfigUpdate(b=bm25_b, k1=bm25_k1),
            cleanupIntervalSeconds=cleanup_interval_seconds,
            stopwords=_StopwordsUpdate(
                preset=stopwords_preset,
                additions=stopwords_additions,
                removals=stopwords_removals,
            ),
        )

    @staticmethod
    def replication(factor: int = 1) -> _ReplicationConfigUpdate:
        """Create a `ReplicationConfigUpdate` object.

        Use this method when defining the `replication_config` argument in `collection.update()`.

        Arguments:
            `factor`
                The replication factor. Defaults to `1`.
        """
        return _ReplicationConfigUpdate(factor=factor)

    @staticmethod
    def vector_index(
        dynamic_ef_factor: Optional[int] = None,
        dynamic_ef_min: Optional[int] = None,
        dynamic_ef_max: Optional[int] = None,
        ef: Optional[int] = None,
        flat_search_cutoff: Optional[int] = None,
        skip: Optional[bool] = None,
        vector_cache_max_objects: Optional[int] = None,
        pq_bit_compression: Optional[bool] = None,
        pq_centroids: Optional[int] = None,
        pq_enabled: Optional[bool] = None,
        pq_encoder_distribution: Optional[PQEncoderDistribution] = None,
        pq_encoder_type: Optional[PQEncoderType] = None,
        pq_segments: Optional[int] = None,
        pq_training_limit: Optional[int] = None,
    ) -> _VectorIndexConfigUpdate:
        """Create a `_VectorIndexConfigUpdate` object.

        Use this method when defining the `vector_index_config` argument in `collection.update()`.

        Arguments:
            See [the docs](https://weaviate.io/developers/weaviate/configuration/indexes#how-to-configure-hnsw) for details!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _VectorIndexConfigUpdate(
            dynamicEfFactor=dynamic_ef_factor,
            dynamicEfMin=dynamic_ef_min,
            dynamicEfMax=dynamic_ef_max,
            ef=ef,
            flatSearchCutoff=flat_search_cutoff,
            skip=skip,
            vectorCacheMaxObjects=vector_cache_max_objects,
            pq=_PQConfigUpdate(
                bitCompression=pq_bit_compression,
                centroids=pq_centroids,
                enabled=pq_enabled,
                encoder=_PQEncoderConfigUpdate(
                    type_=pq_encoder_type,
                    distribution=pq_encoder_distribution,
                ),
                segments=pq_segments,
                trainingLimit=pq_training_limit,
            ),
        )



import sys
from datetime import datetime, timezone
from typing import List, Optional

from pydantic_core._pydantic_core import PydanticUndefined

from weaviate.collections.classes.grpc import MetadataQuery
from weaviate.exceptions import WeaviateAddInvalidPropertyError
from weaviate.types import UUIDS

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated
import pytest as pytest
import uuid

import weaviate
from weaviate.collections.classes.config import (
    Configure,
    PropertyConfig,
)
from weaviate.collections.classes.internal import CrossReference, Reference
from weaviate.collections.classes.orm import BaseProperty, CollectionModelConfig
from weaviate.collections.classes.tenants import Tenant, TenantActivityStatus
from pydantic import Field

REF_TO_UUID = uuid.uuid4()


class Group(BaseProperty):
    name: str


pytest.skip(allow_module_level=True)


@pytest.fixture(scope="module")
def client():
    connection_params = weaviate.ConnectionParams.from_url("http://localhost:8080", 50051)
    client = weaviate.WeaviateClient(connection_params)
    client._collection_model.delete(Group)
    collection = client._collection_model.create(
        CollectionModelConfig[Group](model=Group, vectorizer_config=Configure.Vectorizer.none())
    )
    collection.data.insert(obj=Group(name="Name", uuid=REF_TO_UUID))

    yield client


def test_with_existing_collection(client: weaviate.WeaviateClient):
    obj = client._collection_model.get(Group).data.get_by_id(REF_TO_UUID)
    assert obj.properties.name == "Name"


@pytest.mark.parametrize(
    "member_type,value",
    [
        (str, "1"),
        (int, 1),
        (float, 0.5),
        (List[str], ["1", "2"]),
        (List[int], [1, 2]),
        (List[float], [1.0, 2.1]),
    ],
)
@pytest.mark.parametrize("optional", [True, False])
def test_types(client: weaviate.WeaviateClient, member_type, value, optional: bool):
    if optional:
        member_type = Optional[member_type]

    class ModelTypes(BaseProperty):
        name: member_type

    client._collection_model.delete(ModelTypes)
    collection = client._collection_model.create(
        CollectionModelConfig[ModelTypes](
            model=ModelTypes, vectorizer_config=Configure.Vectorizer.none()
        )
    )
    assert collection.model == ModelTypes

    uuid_object = collection.data.insert(ModelTypes(name=value))
    assert type(uuid_object) is uuid.UUID

    object_get = collection.data.get_by_id(uuid_object)
    assert object_get.properties == ModelTypes(name=value, uuid=uuid_object)

    if optional:
        uuid_object_optional = collection.data.insert(ModelTypes(name=None))
        object_get_optional = collection.data.get_by_id(uuid_object_optional)
        assert object_get_optional.properties == ModelTypes(name=None, uuid=uuid_object_optional)


@pytest.mark.skip(reason="ORM models do not support new references yet")
@pytest.mark.parametrize(
    "member_type, annotation ,value,expected",
    [
        (str, PropertyConfig(index_filterable=False), "value", "text"),
        (UUIDS, CrossReference[Group], [str(REF_TO_UUID)], "Group"),
        (Optional[UUIDS], CrossReference[Group], [str(REF_TO_UUID)], "Group"),
    ],
)
def test_types_annotates(
    client: weaviate.WeaviateClient, member_type, annotation, value, expected: str
):
    class ModelTypes(BaseProperty):
        name: Annotated[member_type, annotation]

    client._collection_model.delete(ModelTypes)
    collection = client._collection_model.create(
        CollectionModelConfig[ModelTypes](
            model=ModelTypes, vectorizer_config=Configure.Vectorizer.none()
        )
    )
    assert collection.model == ModelTypes

    uuid_object = collection.data.insert(ModelTypes(name=value))

    object_get = collection.data.get_by_id(uuid_object)
    assert type(object_get.properties) is ModelTypes

    assert object_get.properties.name == value


def test_create_and_delete(client: weaviate.WeaviateClient):
    class DeleteModel(BaseProperty):
        name: int

    client._collection_model.delete(DeleteModel)
    client._collection_model.create(
        CollectionModelConfig[DeleteModel](
            model=DeleteModel, vectorizer_config=Configure.Vectorizer.none()
        )
    )

    assert client._collection_model.exists(DeleteModel)
    client._collection_model.delete(DeleteModel)
    assert not client._collection_model.exists(DeleteModel)


def test_search(client: weaviate.WeaviateClient):
    class SearchTest(BaseProperty):
        name: str

    client._collection_model.delete(SearchTest)
    collection = client._collection_model.create(
        CollectionModelConfig[SearchTest](
            model=SearchTest, vectorizer_config=Configure.Vectorizer.none()
        )
    )

    collection.data.insert(SearchTest(name="test name"))
    collection.data.insert(SearchTest(name="other words"))

    objects = collection.query.bm25(query="test", return_properties=["name"])
    assert type(objects[0].properties) is SearchTest
    assert objects[0].properties.name == "test name"


def test_tenants(client: weaviate.WeaviateClient):
    class TenantsTest(BaseProperty):
        name: str

    client._collection_model.delete(TenantsTest)
    collection = client._collection_model.create(
        CollectionModelConfig[TenantsTest](
            vectorizer_config=Configure.Vectorizer.none(),
            multi_tenancy_config=Configure.multi_tenancy(enabled=True),
            model=TenantsTest,
        )
    )

    collection.tenants.create([Tenant(name="tenant1")])

    tenants = collection.tenants.get()
    assert len(tenants) == 1
    assert type(tenants["tenant1"]) is Tenant
    assert tenants["tenant1"].name == "tenant1"

    collection.tenants.remove(["tenant1"])

    tenants = collection.tenants.get()
    assert len(tenants) == 0


def test_tenants_activity(client: weaviate.WeaviateClient):
    class TenantsUpdateTest(BaseProperty):
        name: str

    client._collection_model.delete(TenantsUpdateTest)
    collection = client._collection_model.create(
        CollectionModelConfig[TenantsUpdateTest](
            vectorizer_config=Configure.Vectorizer.none(),
            multi_tenancy_config=Configure.multi_tenancy(enabled=True),
            model=TenantsUpdateTest,
        )
    )
    collection.tenants.create(
        [
            Tenant(name="1", activity_status=TenantActivityStatus.HOT),
            Tenant(name="2", activity_status=TenantActivityStatus.COLD),
            Tenant(name="3"),
        ]
    )
    tenants = collection.tenants.get()
    assert tenants["1"].activity_status == TenantActivityStatus.HOT
    assert tenants["2"].activity_status == TenantActivityStatus.COLD
    assert tenants["3"].activity_status == TenantActivityStatus.HOT


def test_tenants_update(client: weaviate.WeaviateClient):
    class TenantsUpdateTest(BaseProperty):
        name: str

    client._collection_model.delete(TenantsUpdateTest)
    collection = client._collection_model.create(
        CollectionModelConfig[TenantsUpdateTest](
            vectorizer_config=Configure.Vectorizer.none(),
            multi_tenancy_config=Configure.multi_tenancy(enabled=True),
            model=TenantsUpdateTest,
        )
    )
    collection.tenants.create([Tenant(name="1", activity_status=TenantActivityStatus.HOT)])
    tenants = collection.tenants.get()
    assert tenants["1"].activity_status == TenantActivityStatus.HOT

    collection.tenants.update([Tenant(name="1", activity_status=TenantActivityStatus.COLD)])
    tenants = collection.tenants.get()
    assert tenants["1"].activity_status == TenantActivityStatus.COLD


def test_multi_searches(client: weaviate.WeaviateClient):
    class TestMultiSearches(BaseProperty):
        name: str

    client._collection_model.delete(TestMultiSearches)
    collection = client._collection_model.create(
        CollectionModelConfig[TestMultiSearches](
            model=TestMultiSearches, vectorizer_config=Configure.Vectorizer.none()
        )
    )

    collection.data.insert(TestMultiSearches(name="some word"))
    collection.data.insert(TestMultiSearches(name="other"))

    objects = collection.query.bm25(
        query="word",
        return_properties=["name"],
        return_metadata=MetadataQuery(last_update_time_unix=True),
    )
    assert objects[0].properties.name == "some word"
    assert objects[0].metadata.last_update_time_unix is not None

    objects = collection.query.bm25(
        query="other", return_properties=["name"], return_metadata=MetadataQuery(uuid=True)
    )
    assert objects[0].properties.name == "other"
    assert objects[0].metadata.uuid is not None
    assert objects[0].metadata.last_update_time_unix is None


@pytest.mark.skip(reason="ORM models do not support references yet")
def test_multi_searches_with_references(client: weaviate.WeaviateClient):
    class TestMultiSearchesWithReferences(BaseProperty):
        name: Optional[str] = None
        group: Optional[CrossReference[Group]] = None

    client._collection_model.delete(TestMultiSearchesWithReferences)
    collection = client._collection_model.create(
        CollectionModelConfig[TestMultiSearchesWithReferences](
            model=TestMultiSearchesWithReferences, vectorizer_config=Configure.Vectorizer.none()
        )
    )

    collection.data.insert(
        TestMultiSearchesWithReferences(name="some word", group=Reference.to(REF_TO_UUID, Group))
    )
    collection.data.insert(
        TestMultiSearchesWithReferences(name="other", group=Reference.to(REF_TO_UUID, Group))
    )

    objects = collection.query.bm25(
        query="word",
        return_properties=["name", "group"],
        return_metadata=MetadataQuery(last_update_time_unix=True),
    )
    assert objects[0].properties.name == "some word"
    assert objects[0].properties.group.objects == []
    assert objects[0].metadata.last_update_time_unix is not None

    objects = collection.query.bm25(
        query="other",
        return_metadata=MetadataQuery(uuid=True),
    )
    assert objects[0].properties.name is None
    assert objects[0].properties.group is None
    assert objects[0].metadata.uuid is not None
    assert objects[0].metadata.last_update_time_unix is None


def test_search_with_tenant(client: weaviate.WeaviateClient):
    class TestTenantSearch(BaseProperty):
        name: str

    client._collection_model.delete(TestTenantSearch)
    collection = client._collection_model.create(
        CollectionModelConfig[TestTenantSearch](
            model=TestTenantSearch,
            vectorizer_config=Configure.Vectorizer.none(),
            multi_tenancy_config=Configure.multi_tenancy(enabled=True),
        )
    )

    collection.tenants.create([Tenant(name="Tenant1"), Tenant(name="Tenant2")])
    tenant1 = collection.with_tenant("Tenant1")
    tenant2 = collection.with_tenant("Tenant2")
    uuid1 = tenant1.data.insert(TestTenantSearch(name="some"))
    objects1 = tenant1.query.bm25(
        query="some", return_properties=["name"], return_metadata=MetadataQuery(uuid=True)
    )
    assert len(objects1) == 1
    assert objects1[0].metadata.uuid == uuid1

    objects2 = tenant2.query.bm25(query="some", return_metadata=MetadataQuery(uuid=True))
    assert len(objects2) == 0


def make_list() -> List[int]:
    return []


@pytest.mark.parametrize(
    "member_type, value_to_add,default,default_factory, exception",
    [
        (int, 10, 5, None, True),
        (int, 10, PydanticUndefined, None, True),
        (Optional[int], 10, PydanticUndefined, None, False),
        (Optional[int], 10, None, None, False),
        (Optional[int], 10, 10, None, False),
        (List[int], [10], None, None, True),
        (Optional[List[int]], [10], None, None, False),
        (List[int], [10], None, make_list, True),
    ],
)
def test_update_properties(
    client: weaviate.WeaviateClient,
    member_type: type,
    value_to_add,
    default,
    default_factory,
    exception: bool,
):
    uuid_first: Optional[uuid.UUID] = None

    def create_original_collection():
        nonlocal uuid_first
        # class definition will be gone when this is out of scope

        class TestPropUpdate(BaseProperty):
            name: str

        client._collection_model.delete(TestPropUpdate)
        collection_first = client._collection_model.create(
            CollectionModelConfig[TestPropUpdate](
                model=TestPropUpdate, vectorizer_config=Configure.Vectorizer.none()
            )
        )
        uuid_first = collection_first.data.insert(TestPropUpdate(name="first"))

    create_original_collection()

    field = Field()
    if default_factory is not None:
        field = Field(default_factory=default_factory)
    elif default is not None:
        field = Field(default=default)

    class TestPropUpdate(BaseProperty):
        name: str
        number: member_type = field

    if exception:
        with pytest.raises(WeaviateAddInvalidPropertyError):
            client._collection_model.update(TestPropUpdate)
    else:
        collection = client._collection_model.update(TestPropUpdate)
        uuid_second = collection.data.insert(TestPropUpdate(name="second", number=value_to_add))
        objects = collection.data.get()
        assert len(objects) == 2
        assert uuid_first is not None
        first = collection.data.get_by_id(uuid_first)

        assert first.properties.name == "first"
        assert (
            first.properties.number == default
            if default is not None and default != PydanticUndefined
            else default_factory()
            if default_factory is not None
            else first.properties.number is None
        )

        second = collection.data.get_by_id(uuid_second)
        assert second.properties.name == "second"
        assert second.properties.number == value_to_add


def test_empty_search_returns_everything(client: weaviate.WeaviateClient):
    class TestReturnEverythingORM(BaseProperty):
        name: Optional[str] = None

    client._collection_model.delete(TestReturnEverythingORM)
    collection = client._collection_model.create(
        CollectionModelConfig[TestReturnEverythingORM](
            model=TestReturnEverythingORM,
            vectorizer_config=Configure.Vectorizer.none(),
        )
    )
    collection.data.insert(TestReturnEverythingORM(name="word"))

    objects = collection.query.bm25(query="word")
    assert objects[0].properties.name is not None
    assert objects[0].properties.name == "word"
    assert objects[0].metadata.uuid is not None
    assert objects[0].metadata.score is not None
    assert objects[0].metadata.last_update_time_unix is not None
    assert objects[0].metadata.creation_time_unix is not None


@pytest.mark.skip(reason="ORM models do not support empty properties in search yet")
def test_empty_return_properties(client: weaviate.WeaviateClient):
    class TestEmptyProperties(BaseProperty):
        name: str

    client._collection_model.delete(TestEmptyProperties)
    collection = client._collection_model.create(
        CollectionModelConfig[TestEmptyProperties](
            model=TestEmptyProperties,
            vectorizer_config=Configure.Vectorizer.none(),
        )
    )
    collection.data.insert(TestEmptyProperties(name="word"))

    objects = collection.query.bm25(query="word", return_metadata=MetadataQuery(uuid=True))
    assert objects[0].properties.name is None


@pytest.mark.skip(reason="ORM models do not support updating reference properties yet")
def test_update_reference_property(client: weaviate.WeaviateClient):
    uuid_first: Optional[uuid.UUID] = None

    def create_original_collection():
        nonlocal uuid_first
        # class definition will be gone when this is out of scope

        class TestRefPropUpdate(BaseProperty):
            name: str

        client._collection_model.delete(TestRefPropUpdate)
        collection_first = client._collection_model.create(
            CollectionModelConfig[TestRefPropUpdate](
                model=TestRefPropUpdate, vectorizer_config=Configure.Vectorizer.none()
            )
        )
        uuid_first = collection_first.data.insert(TestRefPropUpdate(name="first"))

    class TestRefPropUpdate(BaseProperty):
        name: str
        group: CrossReference[Group]

    create_original_collection()
    client._collection_model.update(TestRefPropUpdate)


def test_model_with_datetime_property(client: weaviate.WeaviateClient):
    class TestDatetime(BaseProperty):
        name: str
        date: datetime

    client._collection_model.delete(TestDatetime)
    collection = client._collection_model.create(
        CollectionModelConfig[TestDatetime](
            model=TestDatetime, vectorizer_config=Configure.Vectorizer.none()
        )
    )
    now = datetime.now(timezone.utc)
    collection.data.insert(TestDatetime(name="test", date=now))
    objects = collection.data.get()
    assert len(objects) == 1
    assert objects[0].properties.name == "test"
    assert type(objects[0].properties.date) is datetime

    # assert objects[0].properties.date == now
    # The same issue as in test_collection.py@introduce_date_parsing_to_collections occurs here
    # Weaviate parsing of dates is not perfectly compatible with the python datetime library



import pytest as pytest

import weaviate
from weaviate.collections.classes.config import (
    _CollectionConfig,
    _CollectionConfigSimple,
    Configure,
    ConfigureUpdate,
    Property,
    DataType,
    PQEncoderType,
    PQEncoderDistribution,
    StopwordsPreset,
    VectorDistance,
    _VectorIndexType,
    Vectorizer,
)
from weaviate.collections.classes.tenants import Tenant


@pytest.fixture(scope="module")
def client():
    client = weaviate.connect_to_local(port=8087)
    client.collections.delete_all()
    yield client
    client.collections.delete_all()


def test_collection_list(client: weaviate.WeaviateClient):
    client.collections.create(
        name="TestCollectionList",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="name", data_type=DataType.TEXT),
            Property(name="age", data_type=DataType.INT),
        ],
    )

    collections = client.collections.list_all()
    assert list(collections.keys()) == ["TestCollectionList"]
    assert isinstance(collections["TestCollectionList"], _CollectionConfigSimple)

    collection = client.collections.list_all(False)
    assert list(collection.keys()) == ["TestCollectionList"]
    assert isinstance(collection["TestCollectionList"], _CollectionConfig)

    client.collections.delete("TestCollectionList")


def test_collection_get_simple(client: weaviate.WeaviateClient):
    client.collections.create(
        name="TestCollectionGetSimple",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="name", data_type=DataType.TEXT),
            Property(name="age", data_type=DataType.INT),
        ],
    )

    collection = client.collections.get("TestCollectionGetSimple")
    config = collection.config.get(True)
    assert isinstance(config, _CollectionConfigSimple)

    client.collections.delete("TestCollectionGetSimple")


def test_collection_config_empty(client: weaviate.WeaviateClient):
    collection = client.collections.create(
        name="TestCollectionConfigEmpty",
    )
    config = collection.config.get()

    assert config.name == "TestCollectionConfigEmpty"
    assert config.description is None
    assert config.vectorizer == Vectorizer.NONE

    assert config.properties == []

    assert config.inverted_index_config.bm25.b == 0.75
    assert config.inverted_index_config.bm25.k1 == 1.2
    assert config.inverted_index_config.cleanup_interval_seconds == 60
    assert config.inverted_index_config.index_timestamps is False
    assert config.inverted_index_config.index_property_length is False
    assert config.inverted_index_config.index_null_state is False
    assert config.inverted_index_config.stopwords.additions is None
    assert config.inverted_index_config.stopwords.preset == StopwordsPreset.EN
    assert config.inverted_index_config.stopwords.removals is None

    assert config.multi_tenancy_config.enabled is False

    assert config.replication_config.factor == 1

    assert config.vector_index_config.cleanup_interval_seconds == 300
    assert config.vector_index_config.distance_metric == VectorDistance.COSINE
    assert config.vector_index_config.dynamic_ef_factor == 8
    assert config.vector_index_config.dynamic_ef_max == 500
    assert config.vector_index_config.dynamic_ef_min == 100
    assert config.vector_index_config.ef == -1
    assert config.vector_index_config.ef_construction == 128
    assert config.vector_index_config.flat_search_cutoff == 40000
    assert config.vector_index_config.max_connections == 64
    assert config.vector_index_config.pq.bit_compression is False
    assert config.vector_index_config.pq.centroids == 256
    assert config.vector_index_config.pq.enabled is False
    assert config.vector_index_config.pq.encoder.distribution == PQEncoderDistribution.LOG_NORMAL
    assert config.vector_index_config.pq.encoder.type_ == PQEncoderType.KMEANS
    assert config.vector_index_config.pq.segments == 0
    assert config.vector_index_config.pq.training_limit == 100000
    assert config.vector_index_config.skip is False
    assert config.vector_index_config.vector_cache_max_objects == 1000000000000

    assert config.vector_index_type == _VectorIndexType.HNSW

    client.collections.delete("TestCollectionConfigDefaults")


def test_collection_config_defaults(client: weaviate.WeaviateClient):
    collection = client.collections.create(
        name="TestCollectionConfigDefaults",
        inverted_index_config=Configure.inverted_index(),
        multi_tenancy_config=Configure.multi_tenancy(),
        replication_config=Configure.replication(),
        vector_index_config=Configure.vector_index(),
        vectorizer_config=Configure.Vectorizer.none(),
    )
    config = collection.config.get()

    assert config.name == "TestCollectionConfigDefaults"
    assert config.description is None
    assert config.vectorizer == Vectorizer.NONE

    assert config.properties == []

    assert config.inverted_index_config.bm25.b == 0.75
    assert config.inverted_index_config.bm25.k1 == 1.2
    assert config.inverted_index_config.cleanup_interval_seconds == 60
    assert config.inverted_index_config.index_timestamps is False
    assert config.inverted_index_config.index_property_length is False
    assert config.inverted_index_config.index_null_state is False
    assert config.inverted_index_config.stopwords.additions is None
    assert config.inverted_index_config.stopwords.preset == StopwordsPreset.EN
    assert config.inverted_index_config.stopwords.removals is None

    assert config.multi_tenancy_config.enabled is False

    assert config.replication_config.factor == 1

    assert config.vector_index_config.cleanup_interval_seconds == 300
    assert config.vector_index_config.distance_metric == VectorDistance.COSINE
    assert config.vector_index_config.dynamic_ef_factor == 8
    assert config.vector_index_config.dynamic_ef_max == 500
    assert config.vector_index_config.dynamic_ef_min == 100
    assert config.vector_index_config.ef == -1
    assert config.vector_index_config.ef_construction == 128
    assert config.vector_index_config.flat_search_cutoff == 40000
    assert config.vector_index_config.max_connections == 64
    assert config.vector_index_config.pq.bit_compression is False
    assert config.vector_index_config.pq.centroids == 256
    assert config.vector_index_config.pq.enabled is False
    assert config.vector_index_config.pq.encoder.distribution == PQEncoderDistribution.LOG_NORMAL
    assert config.vector_index_config.pq.encoder.type_ == PQEncoderType.KMEANS
    assert config.vector_index_config.pq.segments == 0
    assert config.vector_index_config.pq.training_limit == 100000
    assert config.vector_index_config.skip is False
    assert config.vector_index_config.vector_cache_max_objects == 1000000000000

    assert config.vector_index_type == _VectorIndexType.HNSW


def test_collection_config_full(client: weaviate.WeaviateClient):
    collection = client.collections.create(
        name="TestCollectionConfigFull",
        description="Test",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="text", data_type=DataType.TEXT),
            Property(name="texts", data_type=DataType.TEXT_ARRAY),
            Property(name="number", data_type=DataType.NUMBER),
            Property(name="numbers", data_type=DataType.NUMBER_ARRAY),
            Property(name="int", data_type=DataType.INT),
            Property(name="ints", data_type=DataType.INT_ARRAY),
            Property(name="date", data_type=DataType.DATE),
            Property(name="dates", data_type=DataType.DATE_ARRAY),
            Property(name="boolean", data_type=DataType.BOOL),
            Property(name="booleans", data_type=DataType.BOOL_ARRAY),
            Property(name="geo", data_type=DataType.GEO_COORDINATES),
            Property(name="phone", data_type=DataType.PHONE_NUMBER),
        ],
        inverted_index_config=Configure.inverted_index(
            bm25_b=0.8,
            bm25_k1=1.3,
            cleanup_interval_seconds=10,
            index_timestamps=True,
            index_property_length=True,
            index_null_state=True,
            stopwords_additions=["a"],
            stopwords_preset=StopwordsPreset.EN,
            stopwords_removals=["the"],
        ),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
        replication_config=Configure.replication(factor=2),
        vector_index_config=Configure.vector_index(
            cleanup_interval_seconds=10,
            distance_metric=VectorDistance.DOT,
            dynamic_ef_factor=6,
            dynamic_ef_max=100,
            dynamic_ef_min=10,
            ef=-2,
            ef_construction=100,
            flat_search_cutoff=41000,
            max_connections=72,
            pq_bit_compression=True,
            pq_centroids=128,
            pq_enabled=True,
            pq_encoder_distribution=PQEncoderDistribution.NORMAL,
            pq_encoder_type=PQEncoderType.TILE,
            pq_segments=4,
            pq_training_limit=1000001,
            skip=True,
            vector_cache_max_objects=100000,
        ),
    )
    config = collection.config.get()

    assert config.name == "TestCollectionConfigFull"
    assert config.description == "Test"
    assert config.vectorizer == Vectorizer.NONE

    assert config.properties[0].name == "text"
    assert config.properties[0].data_type == DataType.TEXT
    assert config.properties[1].name == "texts"
    assert config.properties[1].data_type == DataType.TEXT_ARRAY
    assert config.properties[2].name == "number"
    assert config.properties[2].data_type == DataType.NUMBER
    assert config.properties[3].name == "numbers"
    assert config.properties[3].data_type == DataType.NUMBER_ARRAY
    assert config.properties[4].name == "int"
    assert config.properties[4].data_type == DataType.INT
    assert config.properties[5].name == "ints"
    assert config.properties[5].data_type == DataType.INT_ARRAY
    assert config.properties[6].name == "date"
    assert config.properties[6].data_type == DataType.DATE
    assert config.properties[7].name == "dates"
    assert config.properties[7].data_type == DataType.DATE_ARRAY
    assert config.properties[8].name == "boolean"
    assert config.properties[8].data_type == DataType.BOOL
    assert config.properties[9].name == "booleans"
    assert config.properties[9].data_type == DataType.BOOL_ARRAY
    assert config.properties[10].name == "geo"
    assert config.properties[10].data_type == DataType.GEO_COORDINATES
    assert config.properties[11].name == "phone"
    assert config.properties[11].data_type == DataType.PHONE_NUMBER

    assert config.inverted_index_config.bm25.b == 0.8
    assert config.inverted_index_config.bm25.k1 == 1.3
    assert config.inverted_index_config.cleanup_interval_seconds == 10
    assert config.inverted_index_config.index_timestamps is True
    assert config.inverted_index_config.index_property_length is True
    assert config.inverted_index_config.index_null_state is True
    # assert config.inverted_index_config.stopwords.additions == ["a"] # potential weaviate bug, this returns as None
    assert config.inverted_index_config.stopwords.preset == StopwordsPreset.EN
    assert config.inverted_index_config.stopwords.removals == ["the"]

    assert config.multi_tenancy_config.enabled is True

    assert config.replication_config.factor == 2

    assert config.vector_index_config.cleanup_interval_seconds == 10
    assert config.vector_index_config.distance_metric == VectorDistance.DOT
    assert config.vector_index_config.dynamic_ef_factor == 6
    assert config.vector_index_config.dynamic_ef_max == 100
    assert config.vector_index_config.dynamic_ef_min == 10
    assert config.vector_index_config.ef == -2
    assert config.vector_index_config.ef_construction == 100
    assert config.vector_index_config.flat_search_cutoff == 41000
    assert config.vector_index_config.max_connections == 72
    assert config.vector_index_config.pq.bit_compression is True
    assert config.vector_index_config.pq.centroids == 128
    assert config.vector_index_config.pq.enabled is True
    assert config.vector_index_config.pq.encoder.distribution == PQEncoderDistribution.NORMAL
    # assert config.vector_index_config.pq.encoder.type_ == PQEncoderType.TILE # potential weaviate bug, this returns as PQEncoderType.KMEANS
    assert config.vector_index_config.pq.segments == 4
    assert config.vector_index_config.pq.training_limit == 1000001
    assert config.vector_index_config.skip is True
    assert config.vector_index_config.vector_cache_max_objects == 100000

    assert config.vector_index_type == _VectorIndexType.HNSW

    client.collections.delete("TestCollectionConfigFull")


def test_collection_config_update(client: weaviate.WeaviateClient):
    collection = client.collections.create(
        name="TestCollectionConfigUpdate",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="name", data_type=DataType.TEXT),
            Property(name="age", data_type=DataType.INT),
        ],
    )
    config = collection.config.get()

    assert config.replication_config.factor == 1

    collection.config.update(
        description="Test",
        inverted_index_config=ConfigureUpdate.inverted_index(
            bm25_b=0.8,
            bm25_k1=1.25,
            cleanup_interval_seconds=10,
            stopwords_additions=["a"],
            stopwords_preset=StopwordsPreset.EN,
            stopwords_removals=["the"],
        ),
        replication_config=ConfigureUpdate.replication(factor=2),
        vector_index_config=ConfigureUpdate.vector_index(
            skip=True,
            pq_bit_compression=True,
            pq_centroids=128,
            pq_enabled=True,
            pq_encoder_type=PQEncoderType.TILE,
            pq_encoder_distribution=PQEncoderDistribution.NORMAL,
            pq_segments=4,
            pq_training_limit=100001,
            vector_cache_max_objects=2000000,
        ),
    )

    config = collection.config.get()

    assert config.description == "Test"

    assert config.inverted_index_config.bm25.b == 0.8
    assert config.inverted_index_config.bm25.k1 == 1.25
    assert config.inverted_index_config.cleanup_interval_seconds == 10
    # assert config.inverted_index_config.stopwords.additions is ["a"] # potential weaviate bug, this returns as None
    assert config.inverted_index_config.stopwords.removals == ["the"]

    assert config.replication_config.factor == 2

    assert config.vector_index_config.cleanup_interval_seconds == 300
    assert config.vector_index_config.distance_metric == VectorDistance.COSINE
    assert config.vector_index_config.dynamic_ef_factor == 8
    assert config.vector_index_config.dynamic_ef_max == 500
    assert config.vector_index_config.dynamic_ef_min == 100
    assert config.vector_index_config.ef == -1
    assert config.vector_index_config.ef_construction == 128
    assert config.vector_index_config.flat_search_cutoff == 40000
    assert config.vector_index_config.max_connections == 64
    assert config.vector_index_config.pq.bit_compression is True
    assert config.vector_index_config.pq.centroids == 128
    assert config.vector_index_config.pq.enabled is True
    assert config.vector_index_config.pq.encoder.type_ == PQEncoderType.TILE
    assert config.vector_index_config.pq.encoder.distribution == PQEncoderDistribution.NORMAL
    assert config.vector_index_config.pq.segments == 4
    assert config.vector_index_config.pq.training_limit == 100001
    assert config.vector_index_config.skip is True
    assert config.vector_index_config.vector_cache_max_objects == 2000000

    assert config.vector_index_type == _VectorIndexType.HNSW

    client.collections.delete("TestCollectionSchemaUpdate")


def test_collection_config_get_shards(client: weaviate.WeaviateClient):
    collection = client.collections.create(
        name="TestCollectionConfigGetShards",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="name", data_type=DataType.TEXT),
            Property(name="age", data_type=DataType.INT),
        ],
    )
    shards = collection.config.get_shards()
    assert len(shards)
    assert shards[0].status == "READY"
    assert shards[0].vector_queue_size == 0

    client.collections.delete("TestCollectionConfigGetShards")


def test_collection_config_get_shards_multi_tenancy(client: weaviate.WeaviateClient):
    collection = client.collections.create(
        name="TestCollectionConfigGetShardsMultiTenancy",
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
        properties=[
            Property(name="name", data_type=DataType.TEXT),
            Property(name="age", data_type=DataType.INT),
        ],
    )
    collection.tenants.create([Tenant(name="tenant1"), Tenant(name="tenant2")])
    shards = collection.config.get_shards()
    assert len(shards) == 2

    assert shards[0].status == "READY"
    assert shards[0].vector_queue_size == 0

    assert shards[1].status == "READY"
    assert shards[1].vector_queue_size == 0

    assert "tenant1" in [shard.name for shard in shards]
    assert "tenant2" in [shard.name for shard in shards]

    client.collections.delete("TestCollectionConfigGetShardsMultiTenancy")



import unittest
from unittest.mock import patch, Mock

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.classification.classification import Classification, ConfigBuilder
from weaviate.exceptions import UnexpectedStatusCodeException


class TestClassification(unittest.TestCase):
    def test_schedule(self):
        """
        Test the `schedule` method.
        """

        self.assertIsInstance(Classification(None).schedule(), ConfigBuilder)

    def test_get(self):
        """
        Test the `get` method.
        """

        # error messages
        uuid_type_error = lambda dt: f"'uuid' must be of type str or uuid.UUID, but was: {dt}"
        value_error = "Not valid 'uuid' or 'uuid' can not be extracted from value"
        requests_error_message = "Classification status could not be retrieved."
        unexpected_error_message = "Get classification status"

        # invalid calls
        with self.assertRaises(TypeError) as error:
            Classification(None).get(123)
        check_error_message(self, error, uuid_type_error(int))

        with self.assertRaises(ValueError) as error:
            Classification(None).get("123")
        check_error_message(self, error, value_error)

        mock_conn = mock_connection_func("get", side_effect=RequestsConnectionError("Test!"))
        with self.assertRaises(RequestsConnectionError) as error:
            Classification(mock_conn).get("d087b7c6-a115-5c89-8cb2-f25bdeb9bf92")
        check_error_message(self, error, requests_error_message)

        mock_conn = mock_connection_func("get", status_code=404)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            Classification(mock_conn).get("d087b7c6-a115-5c89-8cb2-f25bdeb9bf92")
        check_startswith_error_message(self, error, unexpected_error_message)

        # valid calls
        mock_conn = mock_connection_func("get", return_json={"OK": "GOOD"}, status_code=200)
        result = Classification(mock_conn).get("d087b7c6-a115-5c89-8cb2-f25bdeb9bf92")
        self.assertEqual(result, {"OK": "GOOD"})

    @patch("weaviate.classification.classification.Classification._check_status")
    def test_is_complete(self, mock_check_status):
        """
        Test the `is_complete` method.
        """

        mock_check_status.return_value = "OK!"
        result = Classification(None).is_complete("Test!")
        self.assertEqual(result, "OK!")
        mock_check_status.assert_called_with("Test!", "completed")

    @patch("weaviate.classification.classification.Classification._check_status")
    def test_is_failed(self, mock_check_status):
        """
        Test the `is_failed` method.
        """

        mock_check_status.return_value = "OK!"
        result = Classification(None).is_failed("Test!")
        self.assertEqual(result, "OK!")
        mock_check_status.assert_called_with("Test!", "failed")

    @patch("weaviate.classification.classification.Classification._check_status")
    def test_is_running(self, mock_check_status):
        """
        Test the `is_running` method.
        """

        mock_check_status.return_value = "OK!"
        result = Classification(None).is_running("Test!")
        self.assertEqual(result, "OK!")
        mock_check_status.assert_called_with("Test!", "running")

    @patch("weaviate.classification.classification.Classification.get")
    def test__check_status(self, mock_get):
        """
        Test the `_check_status` method.
        """

        mock_get.return_value = {"status": "failed"}

        result = Classification(None)._check_status("uuid", "running")
        self.assertFalse(result)

        result = Classification(None)._check_status("uuid", "failed")
        self.assertTrue(result)

        mock_get.side_effect = RequestsConnectionError("Test!")
        result = Classification(None)._check_status("uuid", "running")
        self.assertFalse(result)


class TestConfigBuilder(unittest.TestCase):
    def test_with_type(self):
        """
        Test the `with_type` method.
        """

        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_type("test_type")

        self.assertEqual(config._config, {"type": "test_type"})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_k(self):
        """
        Test the `with_k` method.
        """

        # without `with_settings` called
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_k(4)

        self.assertEqual(config._config, {"settings": {"k": 4}})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

        # with `with_settings` called
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_k(5).with_settings({"test": "OK!"})

        self.assertEqual(config._config, {"settings": {"k": 5, "test": "OK!"}})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_class_name(self):
        """
        Test the `with_class_name` method.
        """

        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        # Correct class name format (capitalized)
        result = config.with_class_name("TestClass")

        self.assertEqual(config._config, {"class": "TestClass"})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

        # Incorrect class name format (capitalized), should be capitalized by the client
        result = config.with_class_name("testClass")

        self.assertEqual(config._config, {"class": "TestClass"})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_classify_properties(self):
        """
        Test the `with_classify_properties` method.
        """

        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_classify_properties(["test1", "test2"])

        self.assertEqual(config._config, {"classifyProperties": ["test1", "test2"]})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_based_on_properties(self):
        """
        Test the `with_based_on_properties` method.
        """

        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_based_on_properties(["test1", "test2"])

        self.assertEqual(config._config, {"basedOnProperties": ["test1", "test2"]})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_source_where_filter(self):
        """
        Test the `with_source_where_filter` method.
        """

        # without other filters set before
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_source_where_filter({"test": "OK!"})

        self.assertEqual(config._config, {"filters": {"sourceWhere": {"test": "OK!"}}})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

        # with other filters set before
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_training_set_where_filter({"test": "OK!"}).with_source_where_filter(
            {"test": "OK!"}
        )

        self.assertEqual(
            config._config,
            {"filters": {"sourceWhere": {"test": "OK!"}, "trainingSetWhere": {"test": "OK!"}}},
        )
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_training_set_where_filter(self):
        """
        Test the `with_training_set_where_filter` method.
        """

        # without other filters set before
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_training_set_where_filter({"test": "OK!"})

        self.assertEqual(config._config, {"filters": {"trainingSetWhere": {"test": "OK!"}}})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

        # with other filters set before
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_target_where_filter({"test": "OK!"}).with_training_set_where_filter(
            {"test": "OK!"}
        )

        self.assertEqual(
            config._config,
            {"filters": {"trainingSetWhere": {"test": "OK!"}, "targetWhere": {"test": "OK!"}}},
        )
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_target_where_filter(self):
        """
        Test the `with_target_where_filter` method.
        """

        # without other filters set before
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_target_where_filter({"test": "OK!"})

        self.assertEqual(config._config, {"filters": {"targetWhere": {"test": "OK!"}}})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

        # with other filters set before
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_source_where_filter({"test": "OK!"}).with_target_where_filter(
            {"test": "OK!"}
        )

        self.assertEqual(
            config._config,
            {"filters": {"targetWhere": {"test": "OK!"}, "sourceWhere": {"test": "OK!"}}},
        )
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_wait_for_completion(self):
        """
        Test the `with_wait_for_completion` method.
        """

        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_wait_for_completion()

        self.assertEqual(config._config, {})
        self.assertTrue(config._wait_for_completion)
        self.assertIs(result, config)

    def test_with_settings(self):
        """
        Test the `with_settings` method.
        """

        # without `with_k` called
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_settings({"test": "OK!"})

        self.assertEqual(config._config, {"settings": {"test": "OK!"}})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

        # with `with_k` called
        config = ConfigBuilder(None, None)
        self.assertEqual(config._config, {})
        self.assertFalse(config._wait_for_completion)

        result = config.with_settings({"test": "OK!"}).with_k(7)

        self.assertEqual(config._config, {"settings": {"k": 7, "test": "OK!"}})
        self.assertFalse(config._wait_for_completion)
        self.assertIs(result, config)

    def test__validate_config(self):
        """
        Test the `_validate_config` method.
        """

        # error messages
        field_error_message = lambda f: f"{f} is not set for this classification"
        settings_error_message = '"settings" should be of type dict'
        k_error_message = "k is not set for this classification"

        # test required fields without "classifyProperties"
        config = (
            ConfigBuilder(None, None)
            .with_type("Test!")
            .with_class_name("Test!")
            .with_based_on_properties(["Test!"])
        )
        with self.assertRaises(ValueError) as error:
            config._validate_config()
        check_error_message(self, error, field_error_message("classifyProperties"))

        # test required fields without "basedOnProperties"
        config = (
            ConfigBuilder(None, None)
            .with_type("Test!")
            .with_class_name("Test!")
            .with_classify_properties(["Test!"])
        )
        with self.assertRaises(ValueError) as error:
            config._validate_config()
        check_error_message(self, error, field_error_message("basedOnProperties"))

        # test required fields without "class"
        config = (
            ConfigBuilder(None, None)
            .with_type("Test!")
            .with_based_on_properties(["Test!"])
            .with_classify_properties(["Test!"])
        )
        with self.assertRaises(ValueError) as error:
            config._validate_config()
        check_error_message(self, error, field_error_message("class"))

        # test required fields without "type"
        config = (
            ConfigBuilder(None, None)
            .with_based_on_properties(["Test!"])
            .with_class_name("Test!")
            .with_classify_properties(["Test!"])
        )
        with self.assertRaises(ValueError) as error:
            config._validate_config()
        check_error_message(self, error, field_error_message("type"))

        # test required fields with all required
        config = (
            ConfigBuilder(None, None)
            .with_based_on_properties(["Test!"])
            .with_type("Test!")
            .with_class_name("Test!")
            .with_classify_properties(["Test!"])
        )
        config._validate_config()

        # test settings
        config = (
            ConfigBuilder(None, None)
            .with_based_on_properties(["Test!"])
            .with_type("Test!")
            .with_class_name("Test!")
            .with_classify_properties(["Test!"])
            .with_settings(["Test!"])
        )
        with self.assertRaises(TypeError) as error:
            config._validate_config()
        check_error_message(self, error, settings_error_message)

        # test knn without k
        config = (
            ConfigBuilder(None, None)
            .with_based_on_properties(["Test!"])
            .with_type("knn")
            .with_class_name("Test!")
            .with_classify_properties(["Test!"])
        )
        with self.assertRaises(ValueError) as error:
            config._validate_config()
        check_error_message(self, error, k_error_message)

        # test knn with k
        config = (
            ConfigBuilder(None, None)
            .with_based_on_properties(["Test!"])
            .with_type("knn")
            .with_class_name("Test!")
            .with_classify_properties(["Test!"])
            .with_k(4)
        )
        config._validate_config()

    def test__start(self):
        """
        Test the `_start` method.
        """

        # error messages
        requests_error_message = "Classification may not started."
        unexpected_error_message = "Start classification"

        # invalid calls
        mock_conn = mock_connection_func("post", side_effect=RequestsConnectionError("Test!"))
        config = ConfigBuilder(mock_conn, None)
        with self.assertRaises(RequestsConnectionError) as error:
            config._start()
        check_error_message(self, error, requests_error_message)
        mock_conn.post.assert_called_with(path="/classifications", weaviate_object={})

        mock_conn = mock_connection_func("post", status_code=200)
        config = ConfigBuilder(mock_conn, None).with_class_name("Test!")
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            config._start()
        check_startswith_error_message(self, error, unexpected_error_message)
        mock_conn.post.assert_called_with(
            path="/classifications", weaviate_object={"class": "Test!"}
        )

        # valid calls
        mock_conn = mock_connection_func("post", status_code=201, return_json="OK!")
        config = ConfigBuilder(mock_conn, None).with_class_name("TestClass").with_type("TestType")
        self.assertEqual(config._start(), "OK!")
        mock_conn.post.assert_called_with(
            path="/classifications", weaviate_object={"class": "TestClass", "type": "TestType"}
        )

    @patch("weaviate.classification.config_builder.ConfigBuilder._start")
    @patch(
        "weaviate.classification.config_builder.ConfigBuilder._validate_config", return_value=None
    )
    def test_do(self, mock_validate_config, mock_start):
        """
        Test the `do` method.
        """

        mock_start.return_value = {"status": "test"}
        config = ConfigBuilder(None, None)
        self.assertEqual(config.do(), {"status": "test"})

        mock_start.return_value = {"status": "test", "id": "test_id"}
        mock_classification = Mock()  # mock self._classification instance

        def mock_waiting(test):
            if mock_waiting.called:
                return False
            mock_waiting.called = True
            return True

        mock_waiting.called = False  # initialize static variable
        mock_classification.is_running.side_effect = mock_waiting
        mock_classification.get.return_value = "test"
        config = ConfigBuilder(None, mock_classification).with_wait_for_completion()
        self.assertEqual(config.do(), "test")

    def test_integration_config(self):
        """
        Test all `with_` methods together that change the configuration.
        """

        config = (
            ConfigBuilder(None, None)
            .with_type("test_type")
            .with_k(4)
            .with_class_name("TestClass")
            .with_classify_properties(["Test1!"])
            .with_based_on_properties(["Test2!"])
            .with_source_where_filter({"test": "OK1!"})
            .with_training_set_where_filter({"test": "OK2!"})
            .with_target_where_filter({"test": "OK3!"})
            .with_settings({"additional": "test_settings"})
        )
        expected_config = {
            "type": "test_type",
            "settings": {"k": 4, "additional": "test_settings"},
            "class": "TestClass",
            "classifyProperties": ["Test1!"],
            "basedOnProperties": ["Test2!"],
            "filters": {
                "sourceWhere": {"test": "OK1!"},
                "trainingSetWhere": {"test": "OK2!"},
                "targetWhere": {"test": "OK3!"},
            },
        }
        self.assertEqual(config._config, expected_config)



import pytest

from weaviate import AuthBearerToken


@pytest.mark.parametrize("expires_in,warning", [(-1, True), (5, False)])
def test_bearer_validation(recwarn, expires_in: int, warning: bool):
    AuthBearerToken(
        access_token="Doesn't matter",
        refresh_token="Doesn't matter",
        expires_in=expires_in,
    )

    if warning:
        assert len(recwarn) == 1
        w = recwarn.pop()
        assert issubclass(w.category, UserWarning)
        assert str(w.message).startswith("Auth003")
    else:
        assert len(recwarn) == 0



from typing import List, Optional

from weaviate.collections.aggregations.base import _Aggregate
from weaviate.collections.classes.aggregate import (
    PropertiesMetrics,
    _AggregateReturn,
    _AggregateGroupByReturn,
)
from weaviate.collections.classes.filters import _Filters


class _OverAll(_Aggregate):
    def over_all(
        self,
        filters: Optional[_Filters] = None,
        limit: Optional[int] = None,
        total_count: bool = False,
        return_metrics: Optional[PropertiesMetrics] = None,
    ) -> _AggregateReturn:
        """Aggregate metrics over all the objects in this collection without any vector search.

        Arguments:
            `filters`
                The filters to apply to the search.
            `limit`
                The maximum number of aggregated objects to return.
            `total_count`
                Whether to include the total number of objects that match the query in the response.
            `return_metrics`
                A list of property metrics to aggregate together after the text search.

        Returns:
            A `_AggregateReturn` object that includes the aggregation objects.

        Raises:
            `weaviate.exceptions.WeaviateQueryException`:
                If an error occurs while performing the query against Weaviate.
        """
        builder = self._base(return_metrics, filters, limit, total_count)
        res = self._do(builder)
        return self._to_aggregate_result(res, return_metrics)


class _OverAllGroupBy(_Aggregate):
    def over_all(
        self,
        group_by: str,
        filters: Optional[_Filters] = None,
        limit: Optional[int] = None,
        total_count: bool = False,
        return_metrics: Optional[PropertiesMetrics] = None,
    ) -> List[_AggregateGroupByReturn]:
        """Aggregate metrics over all the objects in this collection without any vector search grouping the results by a property.

        Arguments:
            `group_by`
                The property name to group the aggregation by.
            `filters`
                The filters to apply to the search.
            `limit`
                The maximum number of aggregated objects to return.
            `total_count`
                Whether to include the total number of objects that match the query in the response.
            `return_metrics`
                A list of property metrics to aggregate together after the text search.

        Returns:
            A `_AggregateGroupByReturn` object that includes the aggregation objects.

        Raises:
            `weaviate.exceptions.WeaviateQueryException`:
                If an error occurs while performing the query against Weaviate.
        """
        builder = (
            self._base(return_metrics, filters, limit, total_count)
            .with_group_by_filter([group_by])
            .with_fields(" groupedBy { path value } ")
        )
        res = self._do(builder)
        return self._to_group_by_result(res, return_metrics)



"""
Test the 'weaviate.batch.requests' functions/classes.
"""
import unittest
from unittest.mock import patch

from test.util import check_error_message
from weaviate.batch.requests import ReferenceBatchRequest, ObjectsBatchRequest


class TestBatchReferences(unittest.TestCase):
    """
    Test the `ReferenceBatchRequest` class.
    """

    @patch("weaviate.batch.requests.get_valid_uuid", side_effect=lambda x: x)
    def test_add_and_get_request_body(self, mock_get_valid_uuid):
        """
        Test the all the ReferenceBatchRequest's methods.
        """

        batch = ReferenceBatchRequest()

        #######################################################################
        # invalid calls
        #######################################################################
        ## error messages
        type_error_message_1 = "'from_object_class_name' argument must be of type str"
        type_error_message_2 = "'from_property_name' argument must be of type str"
        type_error_message_3 = "'to_object_class_name' argument must be of type str"

        with self.assertRaises(TypeError) as error:
            batch.add(10, "some_str", "some_str", "some_str")
        check_error_message(self, error, type_error_message_1)

        with self.assertRaises(TypeError) as error:
            batch.add("some_str", "some_str", True, "some_str")
        check_error_message(self, error, type_error_message_2)

        with self.assertRaises(TypeError) as error:
            batch.add("some_str", "some_str", "some_str", "some_uuid", 1.0)
        check_error_message(self, error, type_error_message_3)

        #######################################################################
        # valid calls
        #######################################################################
        batch = ReferenceBatchRequest()

        #######################################################################
        # test initial values
        self.assertEqual(len(batch), 0)
        self.assertTrue(batch.is_empty())
        self.assertEqual(mock_get_valid_uuid.call_count, 0)

        #######################################################################
        # add first reference
        batch.add("Alpha", "UUID_1", "a", "UUID_2")
        self.assertEqual(len(batch), 1)
        self.assertFalse(batch.is_empty())
        expected_item_1 = {
            "from": "weaviate://localhost/Alpha/UUID_1/a",
            "to": "weaviate://localhost/UUID_2",
        }
        self.assertEqual(batch.get_request_body(), [expected_item_1])
        self.assertEqual(mock_get_valid_uuid.call_count, 2)

        #######################################################################
        # add second reference
        batch.add("Beta", "UUID_2", "b", "UUID_3")
        self.assertEqual(len(batch), 2)
        self.assertFalse(batch.is_empty())
        expected_item_2 = {
            "from": "weaviate://localhost/Beta/UUID_2/b",
            "to": "weaviate://localhost/UUID_3",
        }
        self.assertEqual(batch.get_request_body(), [expected_item_1, expected_item_2])
        self.assertEqual(mock_get_valid_uuid.call_count, 4)

        #######################################################################
        # pop first reference
        self.assertEqual(batch.pop(0), expected_item_1)
        self.assertEqual(len(batch), 1)

        #######################################################################
        # add one reference and pop it pop last reference
        batch.add("Beta", "UUID_3", "b", "UUID_4")
        expected_item_3 = {
            "from": "weaviate://localhost/Beta/UUID_3/b",
            "to": "weaviate://localhost/UUID_4",
        }
        self.assertEqual(len(batch), 2)
        self.assertFalse(batch.is_empty())
        self.assertEqual(batch.pop(), expected_item_3)
        self.assertEqual(len(batch), 1)
        self.assertFalse(batch.is_empty())

        #######################################################################
        # add 2 more references and then empty the batch
        batch.add("Beta", "UUID_4", "b", "UUID_5")
        batch.add("Beta", "UUID_5", "b", "UUID_4")
        self.assertEqual(len(batch), 3)
        self.assertFalse(batch.is_empty())
        batch.empty()
        self.assertEqual(len(batch), 0)
        self.assertTrue(batch.is_empty())


class TestBatchObjects(unittest.TestCase):
    """
    Test the `ObjectsBatchRequest` class.
    """

    @patch("weaviate.batch.requests.uuid4", side_effect=lambda: "d087b7c6a1155c898cb2f25bdeb9bf92")
    @patch("weaviate.batch.requests.get_vector", side_effect=lambda x: x)
    @patch("weaviate.batch.requests.get_valid_uuid", side_effect=lambda x: x)
    def test_add_and_get_request_body(self, mock_get_valid_uuid, mock_get_vector, mock_uuid4):
        """
        Test the all the ObjectsBatchRequest's methods.
        """

        batch = ObjectsBatchRequest()
        #######################################################################
        # invalid calls
        #######################################################################
        ## error messages
        data_type_error_message = "Object must be of type dict"
        class_type_error_message = "Class name must be of type str"

        #######################################################################
        # wrong data_object
        with self.assertRaises(TypeError) as error:
            batch.add(
                data_object=None,
                class_name="Class",
            )
        check_error_message(self, error, data_type_error_message)

        with self.assertRaises(TypeError) as error:
            batch.add(
                data_object=224345,
                class_name="Class",
            )
        check_error_message(self, error, data_type_error_message)

        #######################################################################
        # wrong class_name
        with self.assertRaises(TypeError) as error:
            batch.add(
                data_object={"name": "Optimus Prime"},
                class_name=None,
            )
        check_error_message(self, error, class_type_error_message)

        with self.assertRaises(TypeError) as error:
            batch.add(
                data_object={"name": "Optimus Prime"},
                class_name=["Transformer"],
            )
        check_error_message(self, error, class_type_error_message)

        #######################################################################
        # valid calls
        #######################################################################
        ## test initial values
        self.assertEqual(len(batch), 0)
        self.assertTrue(batch.is_empty())
        self.assertEqual(mock_get_valid_uuid.call_count, 0)
        self.assertEqual(mock_get_vector.call_count, 0)
        expected_return = {"fields": ["ALL"], "objects": []}
        self.assertEqual(batch.get_request_body(), expected_return)

        #######################################################################
        # add an object without 'uuid' and 'vector'
        obj = {"class": "Philosopher", "properties": {"name": "Socrates"}}
        expected_return["objects"].append(
            {
                "class": "Philosopher",
                "properties": {"name": "Socrates"},
                "id": "d087b7c6a1155c898cb2f25bdeb9bf92",
            }
        )
        res_uuid = batch.add(
            data_object=obj["properties"],
            class_name=obj["class"],
        )
        self.assertEqual(len(batch), 1)
        self.assertFalse(batch.is_empty())
        self.assertEqual(mock_get_valid_uuid.call_count, 1)
        self.assertEqual(mock_get_vector.call_count, 0)
        self.assertEqual(batch.get_request_body(), expected_return)
        self.assertEqual(res_uuid, "d087b7c6a1155c898cb2f25bdeb9bf92")
        ## change obj and check if batch does not reflect this change
        obj["properties"]["name"] = "Test"
        self.assertEqual(batch.get_request_body(), expected_return)

        #######################################################################
        # add an object without 'vector'
        obj = {
            "class": "Chemist",
            "properties": {"name": "Marie Curie"},
            "id": "d087b7c6-a115-5c89-8cb2-f25bdeb9bf93",
        }
        expected_return["objects"].append(
            {
                "class": "Chemist",
                "properties": {"name": "Marie Curie"},
                "id": "d087b7c6-a115-5c89-8cb2-f25bdeb9bf93",
            }
        )
        res_uuid = batch.add(
            data_object=obj["properties"],
            class_name=obj["class"],
            uuid=obj["id"],
        )
        self.assertEqual(len(batch), 2)
        self.assertFalse(batch.is_empty())
        self.assertEqual(mock_get_valid_uuid.call_count, 2)
        self.assertEqual(mock_get_vector.call_count, 0)
        self.assertEqual(batch.get_request_body(), expected_return)
        self.assertEqual(res_uuid, "d087b7c6-a115-5c89-8cb2-f25bdeb9bf93")
        ## change obj and check if batch does not reflect this change
        obj["properties"]["name"] = "Test"
        self.assertEqual(batch.get_request_body(), expected_return)

        #######################################################################
        # add an object without 'uuid'
        obj = {"class": "Writer", "properties": {"name": "Stephen King"}, "vector": [1, 2, 3]}
        expected_return["objects"].append(
            {
                "class": "Writer",
                "properties": {"name": "Stephen King"},
                "vector": [1, 2, 3],
                "id": "d087b7c6a1155c898cb2f25bdeb9bf92",
            }
        )
        res_uuid = batch.add(
            data_object=obj["properties"],
            class_name=obj["class"],
            vector=obj["vector"],
        )
        self.assertEqual(len(batch), 3)
        self.assertFalse(batch.is_empty())
        self.assertEqual(mock_get_valid_uuid.call_count, 3)
        self.assertEqual(mock_get_vector.call_count, 1)
        self.assertEqual(batch.get_request_body(), expected_return)
        self.assertEqual(res_uuid, "d087b7c6a1155c898cb2f25bdeb9bf92")
        ## change obj and check if batch does not reflect this change
        obj["properties"]["name"] = "Test"
        self.assertEqual(batch.get_request_body(), expected_return)

        #######################################################################
        # add an object with all arguments
        obj = {
            "class": "Inventor",
            "properties": {"name": "Nikola Tesla"},
            "id": "d087b7c6-a115-5c89-8cb2-f25bdeb9bf95",
            "vector": [1, 2, 3],
        }
        expected_return["objects"].append(
            {
                "class": "Inventor",
                "properties": {"name": "Nikola Tesla"},
                "id": "d087b7c6-a115-5c89-8cb2-f25bdeb9bf95",
                "vector": [1, 2, 3],
            }
        )
        res_uuid = batch.add(
            data_object=obj["properties"],
            class_name=obj["class"],
            uuid=obj["id"],
            vector=obj["vector"],
        )
        self.assertEqual(len(batch), 4)
        self.assertFalse(batch.is_empty())
        self.assertEqual(mock_get_valid_uuid.call_count, 4)
        self.assertEqual(mock_get_vector.call_count, 2)
        self.assertEqual(batch.get_request_body(), expected_return)
        self.assertEqual(res_uuid, "d087b7c6-a115-5c89-8cb2-f25bdeb9bf95")
        ## change obj and check if batch does not reflect this change
        obj["properties"]["name"] = "Test"
        self.assertEqual(batch.get_request_body(), expected_return)

        #######################################################################
        # pop one object with index=1

        self.assertEqual(batch.pop(0), expected_return["objects"][0])
        self.assertEqual(len(batch), 3)
        self.assertFalse(batch.is_empty())
        expected_return["objects"] = expected_return["objects"][1:]

        #######################################################################
        # pop last object

        self.assertEqual(batch.pop(), expected_return["objects"][-1])
        self.assertEqual(len(batch), 2)
        self.assertFalse(batch.is_empty())
        expected_return["objects"] = expected_return["objects"][:-1]

        #######################################################################
        # empty the batch request

        self.assertFalse(batch.is_empty())
        batch.empty()
        self.assertEqual(len(batch), 0)
        self.assertTrue(batch.is_empty())



import json
import re
import time

import pytest
import uuid
from requests import ReadTimeout
from werkzeug.wrappers import Request, Response

import weaviate
from mock_tests.conftest import MOCK_SERVER_URL


def test_no_retry_on_timeout(weaviate_no_auth_mock):
    """Tests that expected timeout exception is raised."""

    def handler(request: Request):
        time.sleep(1.5)  # cause timeout
        return Response(json.dumps({}))

    weaviate_no_auth_mock.expect_request("/v1/batch/objects").respond_with_handler(handler)

    client = weaviate.Client(MOCK_SERVER_URL, timeout_config=(1, 1))

    n = 10
    with pytest.raises(ReadTimeout):
        with client.batch(batch_size=n, timeout_retries=0, dynamic=False) as batch:
            for _ in range(n):
                batch.add_data_object({"name": "test"}, "test", uuid.uuid4())


def test_retry_on_timeout(weaviate_no_auth_mock):
    """Tests that clients resends objects that haven't been added due to a timeout.

    After the timeout, the client checks if
    - An object with the given UUID already exists (using HEAD). Here 50% return that they do NOT exist, eg have to be
    resent.
    - If an object exists, it is checked if the current version in weaviate is identical to the one that is
    sent in the batch. If not, the object in the batch is an update and has to be resent again

    In total 75% are resend.
    """
    added_uuids = []
    first_request = True
    n = 20  # needs to be divisible by 4

    def handler_batch_objects(request: Request):
        nonlocal first_request, n
        if first_request:
            assert len(request.json["objects"]) == n  # all objects are send the first time
            time.sleep(1)  # cause timeout
            first_request = False
        else:
            # 75% of objects have to be resent
            assert len(request.json["objects"]) == n / 4 * 3

        return Response(json.dumps([]))

    weaviate_no_auth_mock.expect_request("/v1/batch/objects").respond_with_handler(
        handler_batch_objects
    )

    # 50% of objects have not been added
    def handler_exists(request: Request):
        if added_uuids.index(request.url.split("/")[-1]) % 2 == 0:
            return Response(json.dumps({}), status=404)
        else:
            return Response(json.dumps({}), status=200)

    weaviate_no_auth_mock.expect_request(
        re.compile("^/v1/objects/Test/"), method="HEAD"
    ).respond_with_handler(handler_exists)

    # 50% of objects are an update to an existing objects and have to be resent
    flip = False

    def handler_get_object(request: Request):
        nonlocal flip
        val = "test" if flip else "other"
        flip = not flip
        return Response(json.dumps({"properties": {"name": val}}))

    weaviate_no_auth_mock.expect_request(
        re.compile("^/v1/objects/Test/"), method="GET"
    ).respond_with_handler(handler_get_object)

    client = weaviate.Client(MOCK_SERVER_URL, timeout_config=(1, 1))
    with client.batch(batch_size=n, timeout_retries=1, dynamic=False) as batch:
        for _ in range(n):
            added_uuids.append(str(uuid.uuid4()))
            batch.add_data_object({"name": "test"}, "test", added_uuids[-1])
    weaviate_no_auth_mock.check_assertions()


def test_retry_on_timeout_all_succesfull(weaviate_no_auth_mock):
    """Test that the client does not resend an empty batch."""
    n = 20

    # handler only responds once => error if a batch is resent
    def handler_batch_objects(request: Request):
        nonlocal n
        assert len(request.json["objects"]) == n  # all objects are send the first time
        time.sleep(1)  # cause timeout
        return Response(json.dumps([]))

    weaviate_no_auth_mock.expect_oneshot_request("/v1/batch/objects").respond_with_handler(
        handler_batch_objects
    )

    # return that all objects are already added successful
    weaviate_no_auth_mock.expect_request(
        re.compile("^/v1/objects/Test/"), method="HEAD"
    ).respond_with_response(Response(json.dumps({}), status=200))
    weaviate_no_auth_mock.expect_request(
        re.compile("^/v1/objects/Test/"), method="GET"
    ).respond_with_json({"properties": {"name": "test"}})

    client = weaviate.Client(MOCK_SERVER_URL, timeout_config=(1, 1))
    with client.batch(batch_size=n, timeout_retries=1, dynamic=False) as batch:
        for _ in range(n):
            batch.add_data_object({"name": "test"}, "test", uuid.uuid4())
    weaviate_no_auth_mock.check_assertions()



from typing import Callable, Generic, Iterable, Iterator, List, Optional, TypeVar
from uuid import UUID

from weaviate.collections.classes.grpc import MetadataQuery
from weaviate.collections.classes.internal import _Object, ReturnProperties
from weaviate.collections.classes.types import Properties


ITERATOR_CACHE_SIZE = 100

P = TypeVar("P")


class _ObjectIterator(Generic[Properties], Iterable[_Object[Properties]]):
    def __init__(
        self,
        fetch_objects_query: Callable[
            [int, Optional[UUID], Optional[MetadataQuery]], List[_Object[Properties]]
        ],
        return_metadata: Optional[MetadataQuery],
        return_properties: Optional[ReturnProperties[Properties]],
    ) -> None:
        self.__query = fetch_objects_query

        self.__meta = return_metadata

        if self.__meta is not None:
            self.__meta.uuid = True
        elif return_properties is not None:
            self.__meta = MetadataQuery(uuid=True)

        self.__iter_object_cache: List[_Object[Properties]] = []
        self.__iter_object_last_uuid: Optional[UUID] = None

    def __iter__(self) -> Iterator[_Object[Properties]]:
        self.__iter_object_cache = []
        self.__iter_object_last_uuid = None
        return self

    def __next__(self) -> _Object[Properties]:
        if len(self.__iter_object_cache) == 0:
            objects = self.__query(
                ITERATOR_CACHE_SIZE,
                self.__iter_object_last_uuid,
                self.__meta,
            )
            self.__iter_object_cache = objects
            if len(self.__iter_object_cache) == 0:
                raise StopIteration

        ret_object = self.__iter_object_cache.pop(0)
        self.__iter_object_last_uuid = ret_object.metadata.uuid
        assert (
            self.__iter_object_last_uuid is not None
        )  # if this is None the iterator will never stop
        return ret_object



import datetime
import uuid as uuid_package
from typing import Dict, Union, List, Tuple

UUID = Union[str, uuid_package.UUID]
UUIDS = Union[List[UUID], UUID]
NUMBER = Union[int, float]
GEO_COORDINATES = Tuple[float, float]

BEACON = "weaviate://localhost/"

PRIMITIVE = Union[str, int, float, bool, datetime.datetime, uuid_package.UUID]

DATATYPE_TO_PYTHON_TYPE = {
    "text": str,
    "int": int,
    "text[]": List[str],
    "int[]": List[int],
    "boolean": bool,
    "boolean[]": List[bool],
    "number": float,
    "number[]": List[float],
    "date": datetime.datetime,
    "date[]": List[datetime.datetime],
    "geoCoordinates": GEO_COORDINATES,
    "object": Dict[str, PRIMITIVE],
    "object[]": List[Dict[str, PRIMITIVE]],
}
PYTHON_TYPE_TO_DATATYPE = {val: key for key, val in DATATYPE_TO_PYTHON_TYPE.items()}
TIME = datetime.datetime

WeaviateField = Union[str, bool, int, float, UUID, GEO_COORDINATES, List["WeaviateField"]]



from http.server import HTTPServer

import pytest as pytest

import weaviate
from mock_tests.conftest import MOCK_SERVER_URL


@pytest.mark.parametrize(
    "version,warning", [("1.16.0", True), ("1.17.2", True), ("1.17.3", False), ("1.18.0", False)]
)
def test_warning_old_weaviate(recwarn, ready_mock: HTTPServer, version: str, warning: bool):
    ready_mock.expect_request("/v1/meta").respond_with_json({"version": version})
    client = weaviate.Client(MOCK_SERVER_URL)

    client.query.get("Class", ["Property"]).with_generate(single_prompt="something")

    if warning:
        assert len(recwarn) == 1
        w = recwarn.pop()
        assert issubclass(w.category, DeprecationWarning)
        assert str(w.message).startswith("Dep003")
    else:
        assert len(recwarn) == 0
